"""Verifiable Credential and Presentation proving methods."""

import asyncio
import logging
import re
from hashlib import sha256
from typing import Any, Optional, Tuple

from anoncreds import (
    AnoncredsError,
    CredentialRevocationState,
    RevocationStatusList,
    W3cCredential,
)

from acapy_agent.anoncreds.registry import AnonCredsRegistry
from acapy_agent.revocation.models.revocation_registry import RevocationRegistry

from ...anoncreds.holder import AnonCredsHolder, AnonCredsHolderError
from ...anoncreds.verifier import AnonCredsVerifier
from ...core.profile import Profile
from ..ld_proofs import LinkedDataProofException, ProofPurpose

LOGGER = logging.getLogger(__name__)


async def create_signed_anoncreds_presentation(
    *,
    profile: Profile,
    pres_definition: dict,
    presentation: dict,
    credentials: list,
    purpose: Optional[ProofPurpose] = None,
    challenge: str,
    domain: Optional[str] = None,
) -> tuple[dict, dict, list]:
    """Sign the presentation with the passed signature suite.

    Will set a default AuthenticationProofPurpose if no proof purpose is passed.

    Args:
        profile (Profile): The profile to use
        pres_definition (dict): The presentation definition
        presentation (dict): The presentation to sign
        credentials (list): The credentials to use for the presentation
        document_loader (DocumentLoader): Document loader to use.
        purpose (ProofPurpose, optional): Purpose to use. Required if challenge is None
        challenge (str, optional): Challenge to use. Required if domain is None.
        domain (str, optional): Domain to use. Only used if purpose is None.
        holder (bool, optional): create a presentation or just the proof request

    Raises:
        LinkedDataProofException: When both purpose and challenge are not provided
            And when signing of the presentation fails

    Returns:
        dict: A verifiable presentation object

    """

    if not challenge:
        raise LinkedDataProofException(
            'A "challenge" param is required when not providing a'
            ' "purpose" (for AuthenticationProofPurpose).'
        )

    w3c_creds = await _load_w3c_credentials(credentials)
    anoncreds_proofrequest, w3c_creds_metadata = await prepare_data_for_presentation(
        presentation, w3c_creds, pres_definition, profile, challenge
    )

    anoncreds_verifier = AnonCredsVerifier(profile)
    (
        schemas,
        cred_defs,
        rev_reg_defs,
        rev_reg_entries,
    ) = await anoncreds_verifier.process_pres_identifiers(w3c_creds_metadata)

    rev_states = await create_rev_states(
        w3c_creds_metadata, rev_reg_defs, rev_reg_entries
    )
    anoncreds_holder = AnonCredsHolder(profile)
    anoncreds_proof = await anoncreds_holder.create_presentation_w3c(
        presentation_request=anoncreds_proofrequest,
        requested_credentials_w3c=w3c_creds,
        credentials_w3c_metadata=w3c_creds_metadata,
        schemas=schemas,
        credential_definitions=cred_defs,
        rev_states=rev_states,
    )

    # TODO any processing to put the returned proof into DIF format
    anoncreds_proof["presentation_submission"] = presentation["presentation_submission"]

    return anoncreds_proof


async def _load_w3c_credentials(credentials: list) -> list:
    """_load_w3c_credentials.

    Args:
        credentials (list): The credentials to load

    Returns:
        list: A list of W3C credentials
    """
    w3c_creds = []
    for credential in credentials:
        try:
            w3c_cred = W3cCredential.load(credential)
            w3c_creds.append(w3c_cred)
        except Exception as err:
            raise LinkedDataProofException(
                "Error loading credential as W3C credential"
            ) from err

    return w3c_creds


async def create_rev_states(
    w3c_creds_metadata: list,
    rev_reg_defs: dict,
    rev_reg_entries: dict,
) -> Optional[dict]:
    """create_rev_states.

    Args:
        profile (Profile): The profile to use
        w3c_creds_metadata (list): The metadata for the credentials
        rev_reg_defs (dict): The revocation registry definitions
        rev_reg_entries (dict): The revocation registry entries

    Returns:
        dict: A dictionary of revocation states
    """
    if not bool(rev_reg_defs and rev_reg_entries):
        return None

    rev_states = {}
    for w3c_cred_cred in w3c_creds_metadata:
        rev_reg_def = rev_reg_defs.get(w3c_cred_cred["rev_reg_id"])
        rev_reg_def["id"] = w3c_cred_cred["rev_reg_id"]
        rev_reg_def_from_registry = RevocationRegistry.from_definition(rev_reg_def, True)
        local_tails_path = await rev_reg_def_from_registry.get_or_fetch_local_tails_path()
        revocation_status_list = RevocationStatusList.load(
            rev_reg_entries.get(w3c_cred_cred["rev_reg_id"])[
                w3c_cred_cred.get("timestamp")
            ]
        )
        rev_reg_index = w3c_cred_cred["rev_reg_index"]
        try:
            rev_state = await asyncio.get_event_loop().run_in_executor(
                None,
                CredentialRevocationState.create,
                rev_reg_def,
                revocation_status_list,
                rev_reg_index,
                local_tails_path,
            )
            rev_states[w3c_cred_cred["rev_reg_id"]] = rev_state
        except AnoncredsError as err:
            raise AnonCredsHolderError("Error creating revocation state") from err

    return rev_states


async def prepare_data_for_presentation(
    presentation: dict,
    w3c_creds: list,
    pres_definition: dict,
    profile: Profile,
    challenge: str,
) -> tuple[dict[str, Any], list, list]:
    """prepare_data_for_presentation.

    Args:
        presentation (dict): The presentation to prepare
        w3c_creds (list): The W3C credentials
        pres_definition (dict): The presentation definition
        profile (Profile): The profile to use
        challenge (str): The challenge to use

    Returns:
        tuple[dict[str, Any], list, list]: A tuple of the anoncreds proof
            request, the W3C credentials metadata, and the W3C credentials
    """

    if not challenge:
        raise LinkedDataProofException("A challenge is required")

    pres_submission = presentation["presentation_submission"]
    descriptor_map = pres_submission["descriptor_map"]
    w3c_creds_metadata = [{} for _ in range(len(w3c_creds))]
    pres_name = (
        pres_definition.get("name") if pres_definition.get("name") else "Proof request"
    )
    challenge_hash = sha256(challenge.encode("utf-8")).hexdigest()
    nonce = str(int(challenge_hash, 16))[:20]

    anoncreds_proofrequest = {
        "version": "1.0",
        "name": pres_name,
        "nonce": nonce,
        "requested_attributes": {},
        "requested_predicates": {},
    }

    for descriptor_map_item in descriptor_map:
        descriptor = next(
            item
            for item in pres_definition["input_descriptors"]
            if item["id"] == descriptor_map_item["id"]
        )

        referent = descriptor_map_item["id"]
        attribute_referent = f"{referent}_attribute"
        predicate_referent_base = f"{referent}_predicate"
        predicate_referent_index = 0

        fields = descriptor["constraints"]["fields"]
        statuses = descriptor["constraints"]["statuses"]

        # descriptor_map_item['path'] should be something
        # like '$.verifiableCredential[n]', we need to extract 'n'
        entry_idx = _extract_cred_idx(descriptor_map_item["path"])
        w3c_cred = w3c_creds[entry_idx]
        schema_id = w3c_cred.schema_id
        cred_def_id = w3c_cred.cred_def_id

        requires_revoc_status = "active" in statuses and statuses["active"][
            "directive"
        ] in ("allowed", "required")

        non_revoked_interval = None
        if requires_revoc_status and w3c_cred.rev_reg_id:
            anoncreds_registry = profile.inject(AnonCredsRegistry)

            result = await anoncreds_registry.get_revocation_list(
                profile, w3c_cred.rev_reg_id, None
            )
            w3c_creds_metadata[entry_idx]["rev_reg_id"] = w3c_cred.rev_reg_id
            w3c_creds_metadata[entry_idx]["timestamp"] = result.revocation_list.timestamp

            non_revoked_interval = {
                "from": result.revocation_list.timestamp,
                "to": result.revocation_list.timestamp,
            }
            w3c_creds_metadata[entry_idx]["rev_reg_index"] = w3c_cred.rev_reg_index
            w3c_creds_metadata[entry_idx]["revoc_status"] = non_revoked_interval

        w3c_creds_metadata[entry_idx]["schema_id"] = schema_id
        w3c_creds_metadata[entry_idx]["cred_def_id"] = cred_def_id
        w3c_creds_metadata[entry_idx]["proof_attrs"] = []
        w3c_creds_metadata[entry_idx]["proof_preds"] = []

        for field in fields:
            path = field["path"][0]

            # check for credential attributes vs other
            if path.startswith("$.credentialSubject."):
                property_name = path.replace("$.credentialSubject.", "")
                if "predicate" in field:
                    # get predicate info
                    pred_filter = field["filter"]
                    (p_type, p_value) = _get_predicate_type_and_value(pred_filter)
                    pred_request = {
                        "name": property_name,
                        "p_type": p_type,
                        "p_value": p_value,
                        "restrictions": [{"cred_def_id": cred_def_id}],
                        "non_revoked": (
                            non_revoked_interval if requires_revoc_status else None
                        ),
                    }
                    predicate_referent = (
                        f"{predicate_referent_base}_{predicate_referent_index}"
                    )
                    predicate_referent_index = predicate_referent_index + 1
                    anoncreds_proofrequest["requested_predicates"][predicate_referent] = (
                        pred_request
                    )
                    w3c_creds_metadata[entry_idx]["proof_preds"].append(
                        predicate_referent
                    )
                else:
                    # no predicate, just a revealed attribute
                    attr_request = {
                        "names": [property_name],
                        "restrictions": [{"cred_def_id": cred_def_id}],
                        "non_revoked": (
                            non_revoked_interval if requires_revoc_status else None
                        ),
                    }
                    # check if we already have this referent ...
                    if (
                        attribute_referent
                        in anoncreds_proofrequest["requested_attributes"]
                    ):
                        anoncreds_proofrequest["requested_attributes"][
                            attribute_referent
                        ]["names"].append(property_name)
                    else:
                        anoncreds_proofrequest["requested_attributes"][
                            attribute_referent
                        ] = attr_request
                        w3c_creds_metadata[entry_idx]["proof_attrs"].append(
                            attribute_referent
                        )
            elif path.endswith(".issuer"):
                # capture issuer - {'path': ['$.issuer'],
                # 'filter': {'type': 'string', 'const': '569XGicsXvYwi512asJkKB'}}
                # TODO prob not a general case
                # issuer_id = field["filter"]["const"]
                pass
            else:
                LOGGER.info("... skipping: %s", path)

    return anoncreds_proofrequest, w3c_creds_metadata


def _extract_cred_idx(item_path: str) -> int:
    """_extract_cred_idx.

    Args:
        item_path (str): path to extract index from

    Raises:
        Exception: No index found in path

    Returns:
        int: extracted index
    """
    match = re.search(r"\[(\d+)\]", item_path)
    if match:
        return int(match.group(1))
    else:
        raise AnonCredsHolderError("No index found in path")


def _get_predicate_type_and_value(pred_filter: dict) -> Tuple[str, str]:
    """_get_predicate_type_and_value.

    Args:
        pred_filter (dict): predicate filter

    Raises:
        Exception: TODO

    Returns:
        Tuple[str, str]: predicate type and value
    """

    supported_properties = {
        "exclusiveMinimum": ">",
        "exclusiveMaximum": "<",
        "minimum": ">=",
        "maximum": "<=",
    }

    # TODO handle multiple predicates?
    for key in pred_filter.keys():
        if key in supported_properties:
            return (supported_properties[key], pred_filter[key])

    # TODO more informative description
    raise AnonCredsHolderError("Unsupported predicate filter")



"""Indy-Credx verifier implementation."""

import asyncio
import logging
from typing import Tuple

from indy_credx import CredxError, Presentation

from ...core.profile import Profile
from ..verifier import IndyVerifier, PresVerifyMsg

LOGGER = logging.getLogger(__name__)


class IndyCredxVerifier(IndyVerifier):
    """Indy-Credx verifier class."""

    def __init__(self, profile: Profile):
        """Initialize an IndyCredxVerifier instance.

        Args:
            profile: an active profile instance

        """
        self.profile = profile

    async def verify_presentation(
        self,
        pres_req,
        pres,
        schemas,
        credential_definitions,
        rev_reg_defs,
        rev_reg_entries,
    ) -> Tuple[bool, list]:
        """Verify a presentation.

        Args:
            pres_req: Presentation request data
            pres: Presentation data
            schemas: Schema data
            credential_definitions: credential definition data
            rev_reg_defs: revocation registry definitions
            rev_reg_entries: revocation registry entries
        """

        accept_legacy_revocation = (
            self.profile.settings.get("revocation.anoncreds_legacy_support", "accept")
            == "accept"
        )
        msgs = []
        try:
            msgs += self.non_revoc_intervals(pres_req, pres, credential_definitions)
            msgs += await self.check_timestamps(
                self.profile, pres_req, pres, rev_reg_defs
            )
            msgs += await self.pre_verify(pres_req, pres)
        except ValueError as err:
            s = str(err)
            msgs.append(f"{PresVerifyMsg.PRES_VALUE_ERROR.value}::{s}")
            LOGGER.error(
                f"Presentation on nonce={pres_req['nonce']} "
                f"cannot be validated (presentation will be marked as Invalid)"
                f": {str(err)}"
            )
            return (False, msgs)

        try:
            presentation = Presentation.load(pres)
            verified = await asyncio.get_event_loop().run_in_executor(
                None,
                presentation.verify,
                pres_req,
                schemas.values(),
                credential_definitions.values(),
                rev_reg_defs.values(),
                rev_reg_entries,
                accept_legacy_revocation,
            )
        except CredxError as err:
            s = str(err)
            msgs.append(f"{PresVerifyMsg.PRES_VERIFY_ERROR.value}::{s}")
            LOGGER.exception(
                f"Validation of presentation on nonce={pres_req['nonce']} "
                "failed with error"
            )
            verified = False

        return (verified, msgs)



TEST_VC_DOCUMENT = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://w3id.org/citizenship/v1",
        "https://w3id.org/security/bbs/v1",
        "https://w3id.org/security/suites/ed25519-2020/v1",
    ],
    "id": "https://issuer.oidp.uscis.gov/credentials/83627465",
    "type": ["VerifiableCredential", "PermanentResidentCard"],
    "issuer": "did:example:489398593",
    "identifier": "83627465",
    "name": "Permanent Resident Card",
    "description": "Government of Example Permanent Resident Card.",
    "issuanceDate": "2019-12-03T12:19:52Z",
    "expirationDate": "2029-12-03T12:19:52Z",
    "credentialSubject": {
        "id": "did:example:b34ca6cd37bbf23",
        "type": ["PermanentResident", "Person"],
        "givenName": "JOHN",
        "familyName": "SMITH",
        "gender": "Male",
        "image": "data:image/png;base64,iVBORw0KGgokJggg==",
        "residentSince": "2015-01-01",
        "lprCategory": "C09",
        "lprNumber": "999-999-999",
        "commuterClassification": "C1",
        "birthCountry": "Bahamas",
        "birthDate": "1958-07-17",
    },
}



import json
from copy import deepcopy
from time import time
from unittest import IsolatedAsyncioTestCase

import pytest

from .....anoncreds.holder import AnonCredsHolder
from .....anoncreds.models.credential_definition import (
    CredDef,
    CredDefValue,
    CredDefValuePrimary,
    GetCredDefResult,
)
from .....anoncreds.models.schema import AnonCredsSchema, GetSchemaResult
from .....anoncreds.registry import AnonCredsRegistry
from .....anoncreds.verifier import AnonCredsVerifier
from .....indy.models.pres_preview import (
    IndyPresAttrSpec,
    IndyPresPredSpec,
    IndyPresPreview,
)
from .....indy.models.xform import indy_proof_req_preview2indy_requested_creds
from .....ledger.base import BaseLedger
from .....ledger.multiple_ledger.ledger_requests_executor import (
    IndyLedgerRequestsExecutor,
)
from .....messaging.decorators.attach_decorator import AttachDecorator
from .....messaging.responder import BaseResponder, MockResponder
from .....multitenant.base import BaseMultitenantManager
from .....multitenant.manager import MultitenantManager
from .....storage.error import StorageNotFoundError
from .....tests import mock
from .....utils.testing import create_test_profile
from .....vc.ld_proofs import DocumentLoader
from .....vc.tests.document_loader import custom_document_loader
from .....vc.vc_ld.validation_result import PresentationVerificationResult
from ...indy import pres_exch_handler as test_indy_util_module
from .. import manager as test_module
from ..formats.dif.handler import DIFPresFormatHandler
from ..formats.dif.tests.test_handler import DIF_PRES
from ..formats.dif.tests.test_handler import DIF_PRES_REQUEST_B as DIF_PRES_REQ
from ..formats.handler import V20PresFormatHandlerError
from ..formats.indy import handler as test_indy_handler
from ..manager import V20PresManager, V20PresManagerError
from ..message_types import ATTACHMENT_FORMAT, PRES_20, PRES_20_PROPOSAL, PRES_20_REQUEST
from ..messages.pres import V20Pres
from ..messages.pres_format import V20PresFormat
from ..messages.pres_problem_report import V20PresProblemReport
from ..messages.pres_proposal import V20PresProposal
from ..messages.pres_request import V20PresRequest
from ..models.pres_exchange import V20PresExRecord

CONN_ID = "connection_id"
ISSUER_DID = "NcYxiDXkpYi6ov5FcYDi1e"
S_ID = f"{ISSUER_DID}:2:vidya:1.0"
CD_ID = f"{ISSUER_DID}:3:CL:{S_ID}:tag1"
RR_ID = f"{ISSUER_DID}:4:{CD_ID}:CL_ACCUM:0"
PROOF_REQ_NAME = "name"
PROOF_REQ_VERSION = "1.0"
PROOF_REQ_NONCE = "12345"

NOW = int(time())
PRES_PREVIEW = IndyPresPreview(
    attributes=[
        IndyPresAttrSpec(name="player", cred_def_id=CD_ID, value="Richie Knucklez"),
        IndyPresAttrSpec(
            name="screenCapture",
            cred_def_id=CD_ID,
            mime_type="image/png",
            value="aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
        ),
    ],
    predicates=[
        IndyPresPredSpec(
            name="highScore", cred_def_id=CD_ID, predicate=">=", threshold=1000000
        )
    ],
)
ANONCREDS_PROOF_REQ_NAME = {
    "name": PROOF_REQ_NAME,
    "version": PROOF_REQ_VERSION,
    "nonce": PROOF_REQ_NONCE,
    "requested_attributes": {
        "0_player_uuid": {
            "name": "player",
            "restrictions": [{"cred_def_id": CD_ID}],
            "non_revoked": {"from": NOW, "to": NOW},
        },
        "0_screencapture_uuid": {
            "name": "screenCapture",
            "restrictions": [{"cred_def_id": CD_ID}],
            "non_revoked": {"from": NOW, "to": NOW},
        },
    },
    "requested_predicates": {
        "0_highscore_GE_uuid": {
            "name": "highScore",
            "p_type": ">=",
            "p_value": 1000000,
            "restrictions": [{"cred_def_id": CD_ID}],
            "non_revoked": {"from": NOW, "to": NOW},
        }
    },
}
ANONCREDS_PROOF_REQ_NAMES = {
    "name": PROOF_REQ_NAME,
    "version": PROOF_REQ_VERSION,
    "nonce": PROOF_REQ_NONCE,
    "requested_attributes": {
        "0_player_uuid": {
            "names": ["player", "screenCapture"],
            "restrictions": [{"cred_def_id": CD_ID}],
            "non_revoked": {"from": NOW, "to": NOW},
        }
    },
    "requested_predicates": {
        "0_highscore_GE_uuid": {
            "name": "highScore",
            "p_type": ">=",
            "p_value": 1000000,
            "restrictions": [{"cred_def_id": CD_ID}],
            "non_revoked": {"from": NOW, "to": NOW},
        }
    },
}
ANONCREDS_PROOF_REQ_SELFIE = {
    "name": PROOF_REQ_NAME,
    "version": PROOF_REQ_VERSION,
    "nonce": PROOF_REQ_NONCE,
    "requested_attributes": {
        "self_player_uuid": {"name": "player"},
        "self_screencapture_uuid": {"name": "screenCapture"},
    },
    "requested_predicates": {
        "0_highscore_GE_uuid": {"name": "highScore", "p_type": ">=", "p_value": 1000000}
    },
}
ANONCREDS_PROOF = {
    "proof": {
        "proofs": [
            {
                "primary_proof": {
                    "eq_proof": {
                        "revealed_attrs": {
                            "player": "51643998292319337989",
                            "screencapture": "124831723185628395682368329568235681",
                        },
                        "a_prime": "98381845469564775640588",
                        "e": "2889201651469315129053056279820725958192110265136",
                        "v": "337782521199137176224",
                        "m": {
                            "master_secret": "88675074759262558623",
                            "date": "3707627155679953691027082306",
                            "highscore": "251972383037120760793174059437326",
                        },
                        "m2": "2892781443118611948331343540849982215419978654911341",
                    },
                    "ge_proofs": [
                        {
                            "u": {
                                "0": "99189584890680947709857922351898933228959",
                                "3": "974568160016086782335901983921278203",
                                "2": "127290395299",
                                "1": "7521808223922",
                            },
                            "r": {
                                "3": "247458",
                                "2": "263046",
                                "1": "285214",
                                "DELTA": "4007402",
                                "0": "12066738",
                            },
                            "mj": "1507606",
                            "alpha": "20251550018805200",
                            "t": {
                                "1": "1262519732727",
                                "3": "82102416",
                                "0": "100578099981822",
                                "2": "47291",
                                "DELTA": "556736142765",
                            },
                            "predicate": {
                                "attr_name": "highscore",
                                "p_type": "GE",
                                "value": 1000000,
                            },
                        }
                    ],
                },
                "non_revoc_proof": {
                    "x_list": {
                        "rho": "128121489ACD4D778ECE",
                        "r": "1890DEFBB8A254",
                        "r_prime": "0A0861FFE96C",
                        "r_prime_prime": "058376CE",
                        "r_prime_prime_prime": "188DF30745A595",
                        "o": "0D0F7FA1",
                        "o_prime": "28165",
                        "m": "0187A9817897FC",
                        "m_prime": "91261D96B",
                        "t": "10FE96",
                        "t_prime": "10856A",
                        "m2": "B136089AAF",
                        "s": "018969A6D",
                        "c": "09186B6A",
                    },
                    "c_list": {
                        "e": "6 1B161",
                        "d": "6 19E861869",
                        "a": "6 541441EE2",
                        "g": "6 7601B068C",
                        "w": "21 10DE6 4 AAAA 5 2458 6 16161",
                        "s": "21 09616 4 1986 5 9797 6 BBBBB",
                        "u": "21 3213123 4 0616FFE 5 323 6 110861861",
                    },
                },
            }
        ],
        "aggregated_proof": {
            "c_hash": "81147637626525127013830996",
            "c_list": [
                [3, 18, 46, 12],
                [3, 136, 2, 39],
                [100, 111, 148, 193],
                [1, 123, 11, 152],
                [2, 138, 162, 227],
                [1, 239, 33, 47],
            ],
        },
    },
    "requested_proof": {
        "revealed_attrs": {
            "0_player_uuid": {
                "sub_proof_index": 0,
                "raw": "Richie Knucklez",
                "encoded": "516439982",
            },
            "0_screencapture_uuid": {
                "sub_proof_index": 0,
                "raw": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                "encoded": "4434954949",
            },
        },
        "self_attested_attrs": {},
        "unrevealed_attrs": {},
        "predicates": {"0_highscore_GE_uuid": {"sub_proof_index": 0}},
    },
    "identifiers": [
        {
            "schema_id": S_ID,
            "cred_def_id": CD_ID,
            "rev_reg_id": RR_ID,
            "timestamp": NOW,
        }
    ],
}
ANONCREDS_PROOF_NAMES = {
    "proof": {
        "proofs": [
            {
                "primary_proof": {
                    "eq_proof": {
                        "revealed_attrs": {
                            "player": "51643998292319337989",
                            "screencapture": "124831723185628395682368329568235681",
                        },
                        "a_prime": "98381845469564775640588",
                        "e": "2889201651469315129053056279820725958192110265136",
                        "v": "337782521199137176224",
                        "m": {
                            "master_secret": "88675074759262558623",
                            "date": "3707627155679953691027082306",
                            "highscore": "251972383037120760793174059437326",
                        },
                        "m2": "2892781443118611948331343540849982215419978654911341",
                    },
                    "ge_proofs": [
                        {
                            "u": {
                                "0": "99189584890680947709857922351898933228959",
                                "3": "974568160016086782335901983921278203",
                                "2": "127290395299",
                                "1": "7521808223922",
                            },
                            "r": {
                                "3": "247458",
                                "2": "263046",
                                "1": "285214",
                                "DELTA": "4007402",
                                "0": "12066738",
                            },
                            "mj": "1507606",
                            "alpha": "20251550018805200",
                            "t": {
                                "1": "1262519732727",
                                "3": "82102416",
                                "0": "100578099981822",
                                "2": "47291",
                                "DELTA": "556736142765",
                            },
                            "predicate": {
                                "attr_name": "highscore",
                                "p_type": "GE",
                                "value": 1000000,
                            },
                        }
                    ],
                },
                "non_revoc_proof": {
                    "x_list": {
                        "rho": "128121489ACD4D778ECE",
                        "r": "1890DEFBB8A254",
                        "r_prime": "0A0861FFE96C",
                        "r_prime_prime": "058376CE",
                        "r_prime_prime_prime": "188DF30745A595",
                        "o": "0D0F7FA1",
                        "o_prime": "28165",
                        "m": "0187A9817897FC",
                        "m_prime": "91261D96B",
                        "t": "10FE96",
                        "t_prime": "10856A",
                        "m2": "B136089AAF",
                        "s": "018969A6D",
                        "c": "09186B6A",
                    },
                    "c_list": {
                        "e": "6 1B161",
                        "d": "6 19E861869",
                        "a": "6 541441EE2",
                        "g": "6 7601B068C",
                        "w": "21 10DE6 4 AAAA 5 2458 6 16161",
                        "s": "21 09616 4 1986 5 9797 6 BBBBB",
                        "u": "21 3213123 4 0616FFE 5 323 6 110861861",
                    },
                },
            }
        ],
        "aggregated_proof": {
            "c_hash": "81147637626525127013830996",
            "c_list": [
                [3, 18, 46, 12],
                [3, 136, 2, 39],
                [100, 111, 148, 193],
                [1, 123, 11, 152],
                [2, 138, 162, 227],
                [1, 239, 33, 47],
            ],
        },
    },
    "requested_proof": {
        "revealed_attrs": {},
        "revealed_attr_groups": {
            "0_player_uuid": {
                "sub_proof_index": 0,
                "values": {
                    "player": {
                        "raw": "Richie Knucklez",
                        "encoded": "516439982",
                    },
                    "screenCapture": {
                        "raw": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                        "encoded": "4434954949",
                    },
                },
            },
        },
        "self_attested_attrs": {},
        "unrevealed_attrs": {},
        "predicates": {"0_highscore_GE_uuid": {"sub_proof_index": 0}},
    },
    "identifiers": [
        {
            "schema_id": S_ID,
            "cred_def_id": CD_ID,
            "rev_reg_id": RR_ID,
            "timestamp": NOW,
        }
    ],
}


class TestV20PresManagerAnonCreds(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.profile.settings.set_value("wallet.type", "askar-anoncreds")
        injector = self.profile.context.injector

        self.ledger = mock.MagicMock(BaseLedger, autospec=True)
        self.ledger.get_schema = mock.CoroutineMock(return_value=mock.MagicMock())
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value={"value": {"revocation": {"...": "..."}}}
        )
        self.ledger.get_revoc_reg_def = mock.CoroutineMock(
            return_value={
                "ver": "1.0",
                "id": RR_ID,
                "revocDefType": "CL_ACCUM",
                "tag": RR_ID.split(":")[-1],
                "credDefId": CD_ID,
                "value": {
                    "IssuanceType": "ISSUANCE_BY_DEFAULT",
                    "maxCredNum": 1000,
                    "publicKeys": {"accumKey": {"z": "1 ..."}},
                    "tailsHash": "3MLjUFQz9x9n5u9rFu8Ba9C5bo4HNFjkPNc54jZPSNaZ",
                    "tailsLocation": "http://sample.ca/path",
                },
            }
        )
        self.ledger.get_revoc_reg_delta = mock.CoroutineMock(
            return_value=(
                {
                    "ver": "1.0",
                    "value": {"prevAccum": "1 ...", "accum": "21 ...", "issued": [1]},
                },
                NOW,
            )
        )
        self.ledger.get_revoc_reg_entry = mock.CoroutineMock(
            return_value=(
                {
                    "ver": "1.0",
                    "value": {"prevAccum": "1 ...", "accum": "21 ...", "issued": [1]},
                },
                NOW,
            )
        )
        injector.bind_instance(BaseLedger, self.ledger)
        injector.bind_instance(
            IndyLedgerRequestsExecutor,
            mock.MagicMock(
                get_ledger_for_identifier=mock.CoroutineMock(
                    return_value=(None, self.ledger)
                )
            ),
        )

        self.holder = mock.MagicMock(AnonCredsHolder, autospec=True)
        get_creds = mock.CoroutineMock(
            return_value=(
                {
                    "cred_info": {
                        "referent": "dummy_reft",
                        "attrs": {
                            "player": "Richie Knucklez",
                            "screenCapture": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                            "highScore": "1234560",
                        },
                    }
                },  # leave this comma: return a tuple
            )
        )
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds
        self.holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "schema_id": S_ID,
                    "cred_def_id": CD_ID,
                    "rev_reg_id": RR_ID,
                    "cred_rev_id": 1,
                }
            )
        )
        self.holder.create_presentation = mock.CoroutineMock(return_value="{}")
        self.holder.create_revocation_state = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "witness": {"omega": "1 ..."},
                    "rev_reg": {"accum": "21 ..."},
                    "timestamp": NOW,
                }
            )
        )
        injector.bind_instance(AnonCredsHolder, self.holder)
        registry = mock.MagicMock(AnonCredsRegistry, autospec=True)
        registry.get_schema = mock.CoroutineMock(
            return_value=GetSchemaResult(
                schema=AnonCredsSchema(
                    issuer_id=ISSUER_DID,
                    name="vidya",
                    version="1.0",
                    attr_names=["player", "screenCapture", "highScore"],
                ),
                schema_id=S_ID,
                resolution_metadata={},
                schema_metadata={},
            )
        )
        registry.get_credential_definition = mock.CoroutineMock(
            return_value=GetCredDefResult(
                credential_definition_id="TUku9MDGa7QALbAJX4oAww:3:CL:531757:MYCO_Consent_Enablement",
                credential_definition=CredDef(
                    issuer_id=ISSUER_DID,
                    schema_id=S_ID,
                    tag="tag",
                    type="CL",
                    value=CredDefValue(
                        primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")
                    ),
                ),
                credential_definition_metadata={},
                resolution_metadata={},
            )
        )
        injector.bind_instance(AnonCredsRegistry, registry)

        self.verifier = mock.MagicMock(AnonCredsVerifier, autospec=True)
        self.verifier.verify_presentation = mock.CoroutineMock(return_value=("true", []))
        injector.bind_instance(AnonCredsVerifier, self.verifier)

        self.manager = V20PresManager(self.profile)

    async def test_record_eq(self):
        same = [
            V20PresExRecord(
                pres_ex_id="dummy-0",
                thread_id="thread-0",
                role=V20PresExRecord.ROLE_PROVER,
            )
        ] * 2
        diff = [
            V20PresExRecord(
                pres_ex_id="dummy-1",
                role=V20PresExRecord.ROLE_PROVER,
            ),
            V20PresExRecord(
                pres_ex_id="dummy-0",
                thread_id="thread-1",
                role=V20PresExRecord.ROLE_PROVER,
            ),
            V20PresExRecord(
                pres_ex_id="dummy-1",
                thread_id="thread-0",
                role=V20PresExRecord.ROLE_VERIFIER,
            ),
        ]

        for i in range(len(same) - 1):
            for j in range(i, len(same)):
                assert same[i] == same[j]

        for i in range(len(diff) - 1):
            for j in range(i, len(diff)):
                assert diff[i] == diff[j] if i == j else diff[i] != diff[j]

    async def test_create_exchange_for_proposal(self):
        proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds", format_=V20PresFormat.Format.ANONCREDS.aries
                )
            ]
        )

        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(V20PresProposal, "serialize", autospec=True),
        ):
            px_rec = await self.manager.create_exchange_for_proposal(
                CONN_ID,
                proposal,
                auto_present=None,
                auto_remove=True,
            )
            save_ex.assert_called_once()

            assert px_rec.thread_id == proposal._thread_id
            assert px_rec.initiator == V20PresExRecord.INITIATOR_SELF
            assert px_rec.role == V20PresExRecord.ROLE_PROVER
            assert px_rec.state == V20PresExRecord.STATE_PROPOSAL_SENT
            assert px_rec.auto_remove is True

    async def test_receive_proposal(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)
        proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds", format_=V20PresFormat.Format.ANONCREDS.aries
                )
            ]
        )
        with mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex:
            px_rec = await self.manager.receive_pres_proposal(
                proposal,
                connection_record,
            )
            save_ex.assert_called_once()

            assert px_rec.state == V20PresExRecord.STATE_PROPOSAL_RECEIVED

    async def test_create_bound_request_a(self):
        comment = "comment"

        proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            proposals_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds")
            ],
        )
        px_rec = V20PresExRecord(
            pres_proposal=proposal.serialize(),
            role=V20PresExRecord.ROLE_VERIFIER,
        )
        px_rec.save = mock.CoroutineMock()
        request_data = {
            "name": PROOF_REQ_NAME,
            "version": PROOF_REQ_VERSION,
            "nonce": PROOF_REQ_NONCE,
        }
        (ret_px_rec, pres_req_msg) = await self.manager.create_bound_request(
            pres_ex_record=px_rec,
            request_data=request_data,
            comment=comment,
        )
        assert ret_px_rec is px_rec
        px_rec.save.assert_called_once()

    async def test_create_bound_request_b(self):
        comment = "comment"

        proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            proposals_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds")
            ],
        )
        px_rec = V20PresExRecord(
            pres_proposal=proposal.serialize(),
            role=V20PresExRecord.ROLE_VERIFIER,
        )
        px_rec.save = mock.CoroutineMock()
        (ret_px_rec, pres_req_msg) = await self.manager.create_bound_request(
            pres_ex_record=px_rec,
            comment=comment,
        )
        assert ret_px_rec is px_rec
        px_rec.save.assert_called_once()

    async def test_create_bound_request_no_format(self):
        px_rec = V20PresExRecord(
            pres_proposal=V20PresProposal(
                formats=[],
                proposals_attach=[],
            ).serialize(),
            role=V20PresExRecord.ROLE_VERIFIER,
        )
        with self.assertRaises(V20PresManagerError) as context:
            await self.manager.create_bound_request(
                pres_ex_record=px_rec,
                request_data={},
                comment="test",
            )
        assert "No supported formats" in str(context.exception)

    async def test_create_pres_no_format(self):
        px_rec = V20PresExRecord(
            pres_proposal=V20PresProposal(
                formats=[],
                proposals_attach=[],
            ).serialize(),
            pres_request=V20PresRequest(
                formats=[], request_presentations_attach=[]
            ).serialize(),
        )
        with self.assertRaises(V20PresManagerError) as context:
            await self.manager.create_pres(
                pres_ex_record=px_rec,
                request_data={},
                comment="test",
            )
        assert "No supported formats" in str(context.exception)

    async def test_create_pres_catch_diferror(self):
        px_rec = V20PresExRecord(
            pres_request=V20PresRequest(
                formats=[
                    V20PresFormat(
                        attach_id="dif",
                        format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                            V20PresFormat.Format.DIF.api
                        ],
                    )
                ],
                request_presentations_attach=[
                    AttachDecorator.data_json(DIF_PRES_REQ, ident="dif")
                ],
            ).serialize(),
        )
        with mock.patch.object(
            DIFPresFormatHandler, "create_pres", autospec=True
        ) as mock_create_pres:
            mock_create_pres.return_value = None
            with self.assertRaises(V20PresManagerError) as context:
                await self.manager.create_pres(
                    pres_ex_record=px_rec,
                    request_data={},
                    comment="test",
                )
            assert "Unable to create presentation. ProblemReport message sent" in str(
                context.exception
            )

    async def test_receive_pres_catch_diferror(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)
        pres_x = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=DIF_PRES,
                    ident="dif",
                )
            ],
        )
        pres_req = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQ, ident="dif")
            ],
        )
        px_rec = V20PresExRecord(
            pres_request=pres_req.serialize(),
            pres=pres_x.serialize(),
        )
        with (
            mock.patch.object(
                DIFPresFormatHandler, "receive_pres", autospec=True
            ) as mock_receive_pres,
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            mock_receive_pres.return_value = False
            retrieve_ex.side_effect = [px_rec]
            with self.assertRaises(V20PresManagerError) as context:
                await self.manager.receive_pres(pres_x, connection_record, None)
            assert "Unable to verify received presentation." in str(context.exception)

    async def test_create_exchange_for_request(self):
        pres_req = V20PresRequest(
            comment="Test",
            will_confirm=True,
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(
                    mapping=ANONCREDS_PROOF_REQ_NAME, ident="anoncreds"
                )
            ],
        )
        pres_req.assign_thread_id("dummy")

        with mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex:
            px_rec = await self.manager.create_exchange_for_request(
                CONN_ID,
                pres_req,
                auto_remove=True,
            )
            save_ex.assert_called_once()

            assert px_rec.thread_id == pres_req._thread_id
            assert px_rec.initiator == V20PresExRecord.INITIATOR_SELF
            assert px_rec.role == V20PresExRecord.ROLE_VERIFIER
            assert px_rec.state == V20PresExRecord.STATE_REQUEST_SENT
            assert px_rec.auto_remove is True

    async def test_receive_pres_request(self):
        px_rec_in = V20PresExRecord()

        with mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex:
            px_rec_out = await self.manager.receive_pres_request(px_rec_in)
            save_ex.assert_called_once()

            assert px_rec_out.state == V20PresExRecord.STATE_REQUEST_RECEIVED

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_create_pres_indy(self):
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds")
            ],
        )
        px_rec_in = V20PresExRecord(pres_request=pres_request.serialize())
        more_magic_rr = mock.MagicMock(
            get_or_fetch_local_tails_path=mock.CoroutineMock(
                return_value="/tmp/sample/tails/path"
            )
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                test_indy_handler, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
            mock.patch.object(
                test_indy_util_module, "RevocationRegistry", autospec=True
            ) as mock_rr,
        ):
            mock_rr.from_definition = mock.MagicMock(return_value=more_magic_rr)

            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )

            req_creds = await indy_proof_req_preview2indy_requested_creds(
                ANONCREDS_PROOF_REQ_NAME, preview=None, holder=self.holder
            )
            request_data = {"anoncreds": req_creds}
            assert not req_creds["self_attested_attributes"]
            assert len(req_creds["requested_attributes"]) == 2
            assert len(req_creds["requested_predicates"]) == 1

            (px_rec_out, pres_msg) = await self.manager.create_pres(
                px_rec_in, request_data
            )
            save_ex.assert_called_once()
            assert px_rec_out.state == V20PresExRecord.STATE_PRESENTATION_SENT

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_create_pres_indy_and_dif(self):
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                ),
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                ),
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds"),
                AttachDecorator.data_json(DIF_PRES_REQ, ident="dif"),
            ],
        )
        px_rec_in = V20PresExRecord(pres_request=pres_request.serialize())
        more_magic_rr = mock.MagicMock(
            get_or_fetch_local_tails_path=mock.CoroutineMock(
                return_value="/tmp/sample/tails/path"
            )
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                test_indy_handler, "AttachDecorator", autospec=True
            ) as mock_attach_decorator_indy,
            mock.patch.object(
                test_indy_util_module, "RevocationRegistry", autospec=True
            ) as mock_rr,
            mock.patch.object(
                DIFPresFormatHandler, "create_pres", autospec=True
            ) as mock_create_pres,
        ):
            mock_rr.from_definition = mock.MagicMock(return_value=more_magic_rr)

            mock_attach_decorator_indy.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator_indy
            )

            mock_create_pres.return_value = (
                PRES_20,
                AttachDecorator.data_json(DIF_PRES, ident="dif"),
            )

            req_creds = await indy_proof_req_preview2indy_requested_creds(
                ANONCREDS_PROOF_REQ_NAME, preview=None, holder=self.holder
            )
            request_data = {"anoncreds": req_creds, "dif": DIF_PRES_REQ}
            assert not req_creds["self_attested_attributes"]
            assert len(req_creds["requested_attributes"]) == 2
            assert len(req_creds["requested_predicates"]) == 1

            (px_rec_out, pres_msg) = await self.manager.create_pres(
                px_rec_in, request_data
            )
            save_ex.assert_called_once()
            assert px_rec_out.state == V20PresExRecord.STATE_PRESENTATION_SENT

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_create_pres_proof_req_non_revoc_interval_none(self):
        indy_proof_req_vcx = deepcopy(ANONCREDS_PROOF_REQ_NAME)
        indy_proof_req_vcx["non_revoked"] = None  # simulate interop with indy-vcx
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(indy_proof_req_vcx, ident="anoncreds")
            ],
        )
        px_rec_in = V20PresExRecord(pres_request=pres_request.serialize())

        more_magic_rr = mock.MagicMock(
            get_or_fetch_local_tails_path=mock.CoroutineMock(
                return_value="/tmp/sample/tails/path"
            )
        )
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with (
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
            ),
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                test_indy_handler, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
            mock.patch.object(
                test_indy_util_module, "RevocationRegistry", autospec=True
            ) as mock_rr,
        ):
            mock_rr.from_definition = mock.MagicMock(return_value=more_magic_rr)

            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )

            req_creds = await indy_proof_req_preview2indy_requested_creds(
                indy_proof_req_vcx, preview=None, holder=self.holder
            )
            request_data = {"anoncreds": req_creds}
            assert not req_creds["self_attested_attributes"]
            assert len(req_creds["requested_attributes"]) == 2
            assert len(req_creds["requested_predicates"]) == 1

            (px_rec_out, pres_msg) = await self.manager.create_pres(
                px_rec_in, request_data
            )
            save_ex.assert_called_once()
            assert px_rec_out.state == V20PresExRecord.STATE_PRESENTATION_SENT

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_create_pres_self_asserted(self):
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_SELFIE, ident="anoncreds")
            ],
        )
        px_rec_in = V20PresExRecord(pres_request=pres_request.serialize())

        more_magic_rr = mock.MagicMock(
            get_or_fetch_local_tails_path=mock.CoroutineMock(
                return_value="/tmp/sample/tails/path"
            )
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                test_indy_handler, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
            mock.patch.object(
                test_indy_util_module, "RevocationRegistry", autospec=True
            ) as mock_rr,
        ):
            mock_rr.from_definition = mock.MagicMock(return_value=more_magic_rr)

            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )

            req_creds = await indy_proof_req_preview2indy_requested_creds(
                ANONCREDS_PROOF_REQ_SELFIE, preview=None, holder=self.holder
            )
            request_data = {"anoncreds": req_creds}

            assert len(req_creds["self_attested_attributes"]) == 3
            assert not req_creds["requested_attributes"]
            assert not req_creds["requested_predicates"]

            (px_rec_out, pres_msg) = await self.manager.create_pres(
                px_rec_in, request_data
            )
            save_ex.assert_called_once()
            assert px_rec_out.state == V20PresExRecord.STATE_PRESENTATION_SENT

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_create_pres_no_revocation(self):
        self.ledger = mock.MagicMock(BaseLedger, autospec=True)
        self.ledger.get_schema = mock.CoroutineMock(return_value=mock.MagicMock())
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value={"value": {"revocation": None}}
        )
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)

        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds")
            ],
        )
        px_rec_in = V20PresExRecord(pres_request=pres_request.serialize())

        self.holder = mock.MagicMock(AnonCredsHolder, autospec=True)
        get_creds = mock.CoroutineMock(
            return_value=(
                {
                    "cred_info": {"referent": "dummy_reft"},
                    "attrs": {
                        "player": "Richie Knucklez",
                        "screenCapture": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                        "highScore": "1234560",
                    },
                },  # leave this comma: return a tuple
            )
        )
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds
        self.holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "schema_id": S_ID,
                    "cred_def_id": CD_ID,
                    "rev_reg_id": None,
                    "cred_rev_id": None,
                }
            )
        )
        self.holder.create_presentation = mock.AsyncMock(return_value="{}")
        self.profile.context.injector.bind_instance(AnonCredsHolder, self.holder)

        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                test_indy_handler, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
            mock.patch.object(
                test_indy_util_module.LOGGER, "info", mock.MagicMock()
            ) as mock_log_info,
        ):
            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )

            req_creds = await indy_proof_req_preview2indy_requested_creds(
                ANONCREDS_PROOF_REQ_NAME, preview=None, holder=self.holder
            )
            request_data = {
                "anoncreds": {
                    "self_attested_attributes": req_creds["self_attested_attributes"],
                    "requested_attributes": req_creds["requested_attributes"],
                    "requested_predicates": req_creds["requested_predicates"],
                }
            }

            (px_rec_out, pres_msg) = await self.manager.create_pres(
                px_rec_in, request_data
            )
            save_ex.assert_called_once()
            assert px_rec_out.state == V20PresExRecord.STATE_PRESENTATION_SENT

            # exercise superfluous timestamp removal
            for pred_reft_spec in req_creds["requested_predicates"].values():
                pred_reft_spec["timestamp"] = 1234567890
            request_data = {
                "anoncreds": {
                    "self_attested_attributes": req_creds["self_attested_attributes"],
                    "requested_attributes": req_creds["requested_attributes"],
                    "requested_predicates": req_creds["requested_predicates"],
                }
            }
            await self.manager.create_pres(px_rec_in, request_data)
            mock_log_info.assert_called_once()

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_create_pres_bad_revoc_state(self):
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds")
            ],
        )
        px_rec_in = V20PresExRecord(pres_request=pres_request.serialize())

        self.holder = mock.MagicMock(AnonCredsHolder, autospec=True)
        get_creds = mock.CoroutineMock(
            return_value=(
                {
                    "cred_info": {"referent": "dummy_reft"},
                    "attrs": {
                        "player": "Richie Knucklez",
                        "screenCapture": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                        "highScore": "1234560",
                    },
                },  # leave this comma: return a tuple
            )
        )
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds

        self.holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "schema_id": S_ID,
                    "cred_def_id": CD_ID,
                    "rev_reg_id": RR_ID,
                    "cred_rev_id": 1,
                }
            )
        )
        self.holder.create_presentation = mock.AsyncMock(return_value="{}")
        self.holder.create_revocation_state = mock.AsyncMock(
            side_effect=test_indy_util_module.AnonCredsHolderError(
                "Problem", {"message": "Nope"}
            )
        )
        self.profile.context.injector.bind_instance(AnonCredsHolder, self.holder)

        more_magic_rr = mock.MagicMock(
            get_or_fetch_local_tails_path=mock.CoroutineMock(
                return_value="/tmp/sample/tails/path"
            )
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True),
            mock.patch.object(
                test_indy_handler, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
            mock.patch.object(
                test_indy_util_module, "RevocationRegistry", autospec=True
            ) as mock_rr,
            mock.patch.object(test_indy_util_module.LOGGER, "error", mock.MagicMock()),
        ):
            mock_rr.from_definition = mock.MagicMock(return_value=more_magic_rr)

            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )
            request_data = {}
            with self.assertRaises(test_indy_util_module.AnonCredsHolderError):
                await self.manager.create_pres(px_rec_in, request_data)

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_create_pres_multi_matching_proposal_creds_names(self):
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAMES, ident="anoncreds")
            ],
        )
        px_rec_in = V20PresExRecord(pres_request=pres_request.serialize())

        self.holder = mock.MagicMock(AnonCredsHolder, autospec=True)
        get_creds = mock.CoroutineMock(
            return_value=(
                {
                    "cred_info": {
                        "referent": "dummy_reft_0",
                        "cred_def_id": CD_ID,
                        "attrs": {
                            "player": "Richie Knucklez",
                            "screenCapture": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                            "highScore": "1234560",
                        },
                    }
                },
                {
                    "cred_info": {
                        "referent": "dummy_reft_1",
                        "cred_def_id": CD_ID,
                        "attrs": {
                            "player": "Richie Knucklez",
                            "screenCapture": "aW1hZ2luZSBhbm90aGVyIHNjcmVlbiBjYXB0dXJl",
                            "highScore": "1515880",
                        },
                    }
                },
            )
        )
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds
        self.holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "schema_id": S_ID,
                    "cred_def_id": CD_ID,
                    "rev_reg_id": RR_ID,
                    "cred_rev_id": 1,
                }
            )
        )
        self.holder.create_presentation = mock.CoroutineMock(return_value="{}")
        self.holder.create_revocation_state = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "witness": {"omega": "1 ..."},
                    "rev_reg": {"accum": "21 ..."},
                    "timestamp": NOW,
                }
            )
        )
        self.profile.context.injector.bind_instance(AnonCredsHolder, self.holder)

        more_magic_rr = mock.MagicMock(
            get_or_fetch_local_tails_path=mock.CoroutineMock(
                return_value="/tmp/sample/tails/path"
            )
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                test_indy_handler, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
            mock.patch.object(
                test_indy_util_module, "RevocationRegistry", autospec=True
            ) as mock_rr,
        ):
            mock_rr.from_definition = mock.MagicMock(return_value=more_magic_rr)

            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )

            req_creds = await indy_proof_req_preview2indy_requested_creds(
                ANONCREDS_PROOF_REQ_NAMES, preview=None, holder=self.holder
            )
            assert not req_creds["self_attested_attributes"]
            assert len(req_creds["requested_attributes"]) == 1
            assert len(req_creds["requested_predicates"]) == 1
            request_data = {"anoncreds": req_creds}
            (px_rec_out, pres_msg) = await self.manager.create_pres(
                px_rec_in, request_data
            )
            save_ex.assert_called_once()
            assert px_rec_out.state == V20PresExRecord.STATE_PRESENTATION_SENT

    async def test_no_matching_creds_for_proof_req(self):
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAMES, ident="anoncreds")
            ],
        )
        V20PresExRecord(pres_request=pres_request.serialize())
        get_creds = mock.CoroutineMock(return_value=())
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds

        with self.assertRaises(ValueError):
            await indy_proof_req_preview2indy_requested_creds(
                ANONCREDS_PROOF_REQ_NAMES, preview=None, holder=self.holder
            )

        get_creds = mock.CoroutineMock(
            return_value=(
                {
                    "cred_info": {"referent": "dummy_reft"},
                    "attrs": {
                        "player": "Richie Knucklez",
                        "screenCapture": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                        "highScore": "1234560",
                    },
                },  # leave this comma: return a tuple
            )
        )
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds
        await indy_proof_req_preview2indy_requested_creds(
            ANONCREDS_PROOF_REQ_NAMES, preview=None, holder=self.holder
        )

    async def test_no_matching_creds_indy_handler(self):
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAMES, ident="anoncreds")
            ],
        )
        px_rec_in = V20PresExRecord(pres_request=pres_request.serialize())
        get_creds = mock.CoroutineMock(return_value=())
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds

        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True),
            mock.patch.object(
                test_indy_handler, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
        ):
            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )
            request_data = {}
            with self.assertRaises(
                test_indy_handler.V20PresFormatHandlerError
            ) as context:
                (px_rec_out, pres_msg) = await self.manager.create_pres(
                    px_rec_in, request_data
                )
            assert "AnonCreds interface requires AskarAnonCreds profile" in str(
                context.exception
            )

    async def test_receive_pres(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            proposals_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds")
            ],
        )
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds")
            ],
        )
        pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF, ident="anoncreds")
            ],
        )
        pres.assign_thread_id("thread-id")

        px_rec_dummy = V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            pres_request=pres_request.serialize(),
        )

        # cover by_format property
        by_format = px_rec_dummy.by_format

        assert by_format.get("pres_proposal").get("anoncreds") == ANONCREDS_PROOF_REQ_NAME
        assert by_format.get("pres_request").get("anoncreds") == ANONCREDS_PROOF_REQ_NAME

        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.side_effect = [px_rec_dummy]
            px_rec_out = await self.manager.receive_pres(pres, connection_record, None)
            assert retrieve_ex.call_args.args[1] == {"thread_id": "thread-id"}
            assert retrieve_ex.call_args.args[2] == {
                "role": V20PresExRecord.ROLE_VERIFIER,
                "connection_id": CONN_ID,
            }
            save_ex.assert_called_once()
            assert px_rec_out.state == (V20PresExRecord.STATE_PRESENTATION_RECEIVED)

    async def test_receive_pres_receive_pred_value_mismatch_punt_to_indy(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            proposals_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds")
            ],
        )
        indy_proof_req = deepcopy(ANONCREDS_PROOF_REQ_NAME)
        indy_proof_req["requested_predicates"]["0_highscore_GE_uuid"]["restrictions"][0][
            "attr::player::value"
        ] = "impostor"
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF, ident="anoncreds")
            ],
        )
        pres.assign_thread_id("thread-id")

        px_rec_dummy = V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            pres_request=pres_request.serialize(),
        )

        # cover by_format property
        by_format = px_rec_dummy.by_format

        assert by_format.get("pres_proposal").get("anoncreds") == ANONCREDS_PROOF_REQ_NAME
        assert by_format.get("pres_request").get("anoncreds") == indy_proof_req

        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.side_effect = [px_rec_dummy]
            px_rec_out = await self.manager.receive_pres(pres, connection_record, None)
            assert retrieve_ex.call_args.args[1] == {"thread_id": "thread-id"}
            assert retrieve_ex.call_args.args[2] == {
                "role": V20PresExRecord.ROLE_VERIFIER,
                "connection_id": CONN_ID,
            }
            save_ex.assert_called_once()
            assert px_rec_out.state == (V20PresExRecord.STATE_PRESENTATION_RECEIVED)

    async def test_receive_pres_indy_no_predicate_restrictions(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)
        indy_proof_req = {
            "name": PROOF_REQ_NAME,
            "version": PROOF_REQ_VERSION,
            "nonce": PROOF_REQ_NONCE,
            "requested_attributes": {
                "0_player_uuid": {
                    "name": "player",
                    "restrictions": [{"cred_def_id": CD_ID}],
                    "non_revoked": {"from": NOW, "to": NOW},
                },
                "0_screencapture_uuid": {
                    "name": "screenCapture",
                    "restrictions": [{"cred_def_id": CD_ID}],
                    "non_revoked": {"from": NOW, "to": NOW},
                },
            },
            "requested_predicates": {
                "0_highscore_GE_uuid": {
                    "name": "highScore",
                    "p_type": ">=",
                    "p_value": 1000000,
                    "restrictions": [],
                    "non_revoked": {"from": NOW, "to": NOW},
                }
            },
        }
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF, ident="anoncreds")
            ],
        )
        pres.assign_thread_id("thread-id")

        px_rec_dummy = V20PresExRecord(
            pres_request=pres_request.serialize(),
        )

        # cover by_format property
        by_format = px_rec_dummy.by_format

        assert by_format.get("pres_request").get("anoncreds") == indy_proof_req

        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.side_effect = [px_rec_dummy]
            px_rec_out = await self.manager.receive_pres(pres, connection_record, None)
            assert retrieve_ex.call_args.args[1] == {"thread_id": "thread-id"}
            assert retrieve_ex.call_args.args[2] == {
                "role": V20PresExRecord.ROLE_VERIFIER,
                "connection_id": CONN_ID,
            }
            save_ex.assert_called_once()
            assert px_rec_out.state == (V20PresExRecord.STATE_PRESENTATION_RECEIVED)

    async def test_receive_pres_indy_no_attr_restrictions(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)
        indy_proof_req = {
            "name": PROOF_REQ_NAME,
            "version": PROOF_REQ_VERSION,
            "nonce": PROOF_REQ_NONCE,
            "requested_attributes": {
                "0_player_uuid": {
                    "name": "player",
                    "restrictions": [],
                    "non_revoked": {"from": NOW, "to": NOW},
                }
            },
            "requested_predicates": {},
        }
        proof = deepcopy(ANONCREDS_PROOF)
        proof["requested_proof"]["revealed_attrs"] = {
            "0_player_uuid": {
                "sub_proof_index": 0,
                "raw": "Richie Knucklez",
                "encoded": "516439982",
            }
        }
        proof["requested_proof"]["predicates"] = {}
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[AttachDecorator.data_base64(proof, ident="anoncreds")],
        )
        pres.assign_thread_id("thread-id")

        px_rec_dummy = V20PresExRecord(
            pres_request=pres_request.serialize(),
        )

        # cover by_format property
        by_format = px_rec_dummy.by_format

        assert by_format.get("pres_request").get("anoncreds") == indy_proof_req

        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.side_effect = [px_rec_dummy]
            px_rec_out = await self.manager.receive_pres(pres, connection_record, None)
            assert retrieve_ex.call_args.args[1] == {"thread_id": "thread-id"}
            assert retrieve_ex.call_args.args[2] == {
                "role": V20PresExRecord.ROLE_VERIFIER,
                "connection_id": CONN_ID,
            }
            save_ex.assert_called_once()
            assert px_rec_out.state == (V20PresExRecord.STATE_PRESENTATION_RECEIVED)

    async def test_receive_pres_bait_and_switch_attr_name(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)
        indy_proof_req = deepcopy(ANONCREDS_PROOF_REQ_NAME)
        indy_proof_req["requested_attributes"]["0_screencapture_uuid"]["restrictions"][0][
            "attr::screenCapture::value"
        ] = "c2NyZWVuIGNhcHR1cmUgc2hvd2luZyBzY29yZSBpbiB0aGUgbWlsbGlvbnM="
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            proposals_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds")
            ],
        )
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_x = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF, ident="anoncreds")
            ],
        )
        px_rec_dummy = V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            pres_request=pres_request.serialize(),
            pres=pres_x.serialize(),
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True),
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = px_rec_dummy
            with self.assertRaises(V20PresFormatHandlerError) as context:
                await self.manager.receive_pres(pres_x, connection_record, None)
            assert "does not satisfy proof request restrictions" in str(context.exception)

        indy_proof_req["requested_attributes"]["shenanigans"] = indy_proof_req[
            "requested_attributes"
        ].pop("0_screencapture_uuid")
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            proposals_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_x = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF, ident="anoncreds")
            ],
        )

        px_rec_dummy = V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            pres_request=pres_request.serialize(),
            pres=pres_x.serialize(),
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True),
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = px_rec_dummy
            with self.assertRaises(V20PresFormatHandlerError) as context:
                await self.manager.receive_pres(pres_x, connection_record, None)
            assert "Presentation referent" in str(context.exception)

    async def test_receive_pres_bait_and_switch_attr_names(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)
        indy_proof_req = deepcopy(ANONCREDS_PROOF_REQ_NAMES)
        indy_proof_req["requested_attributes"]["0_player_uuid"]["restrictions"][0][
            "attr::screenCapture::value"
        ] = "c2NyZWVuIGNhcHR1cmUgc2hvd2luZyBzY29yZSBpbiB0aGUgbWlsbGlvbnM="
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            proposals_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_x = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_NAMES, ident="anoncreds")
            ],
        )

        px_rec_dummy = V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            pres_request=pres_request.serialize(),
            pres=pres_x.serialize(),
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True),
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = px_rec_dummy
            with self.assertRaises(V20PresFormatHandlerError) as context:
                await self.manager.receive_pres(pres_x, connection_record, None)
            assert "does not satisfy proof request restrictions " in str(
                context.exception
            )

        indy_proof_req["requested_attributes"]["shenanigans"] = indy_proof_req[
            "requested_attributes"
        ].pop("0_player_uuid")
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            proposals_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_x = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_NAMES, ident="anoncreds")
            ],
        )

        px_rec_dummy = V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            pres_request=pres_request.serialize(),
            pres=pres_x.serialize(),
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True),
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = px_rec_dummy
            with self.assertRaises(V20PresFormatHandlerError) as context:
                await self.manager.receive_pres(pres_x, connection_record, None)
            assert "Presentation referent" in str(context.exception)

    async def test_receive_pres_bait_and_switch_pred(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)
        indy_proof_req = deepcopy(ANONCREDS_PROOF_REQ_NAME)
        indy_proof_req["requested_predicates"] = {}
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            proposals_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds")
            ],
        )
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_x = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF, ident="anoncreds")
            ],
        )

        px_rec_dummy = V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            pres_request=pres_request.serialize(),
            pres=pres_x.serialize(),
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True),
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = px_rec_dummy
            with self.assertRaises(V20PresFormatHandlerError) as context:
                await self.manager.receive_pres(pres_x, connection_record, None)
            assert "not in proposal request" in str(context.exception)

        indy_proof_req["requested_predicates"]["0_highscore_GE_uuid"] = {
            "name": "shenanigans",
            "p_type": ">=",
            "p_value": 1000000,
            "restrictions": [{"cred_def_id": CD_ID}],
            "non_revoked": {"from": NOW, "to": NOW},
        }
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            proposals_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_x = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF, ident="anoncreds")
            ],
        )

        px_rec_dummy = V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            pres_request=pres_request.serialize(),
            pres=pres_x.serialize(),
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True),
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = px_rec_dummy
            with self.assertRaises(V20PresFormatHandlerError) as context:
                await self.manager.receive_pres(pres_x, connection_record, None)
            assert "shenanigans not in presentation" in str(context.exception)

        indy_proof_req["requested_predicates"]["0_highscore_GE_uuid"] = {
            "name": "highScore",
            "p_type": ">=",
            "p_value": 8000000,  # propose >= 8 million, prove >= 1 million
            "restrictions": [{"cred_def_id": CD_ID}],
            "non_revoked": {"from": NOW, "to": NOW},
        }
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            proposals_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_x = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF, ident="anoncreds")
            ],
        )

        px_rec_dummy = V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            pres_request=pres_request.serialize(),
            pres=pres_x.serialize(),
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True),
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = px_rec_dummy
            with self.assertRaises(V20PresFormatHandlerError) as context:
                await self.manager.receive_pres(pres_x, connection_record, None)
            assert "highScore mismatches proposal request" in str(context.exception)

        indy_proof_req["requested_predicates"]["0_highscore_GE_uuid"] = {
            "name": "highScore",
            "p_type": ">=",
            "p_value": 1000000,
            "restrictions": [{"issuer_did": "FFFFFFFFFFFFFFFFFFFFFF"}],  # fake issuer
            "non_revoked": {"from": NOW, "to": NOW},
        }
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            proposals_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_base64(indy_proof_req, ident="anoncreds")
            ],
        )
        pres_x = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF, ident="anoncreds")
            ],
        )

        px_rec_dummy = V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            pres_request=pres_request.serialize(),
            pres=pres_x.serialize(),
        )
        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True),
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = px_rec_dummy
            with self.assertRaises(V20PresFormatHandlerError) as context:
                await self.manager.receive_pres(pres_x, connection_record, None)
            assert "does not satisfy proof request restrictions " in str(
                context.exception
            )

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_verify_pres(self):
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            will_confirm=True,
            request_presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds")
            ],
        )
        pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF, ident="anoncreds")
            ],
        )
        px_rec_in = V20PresExRecord(
            pres_request=pres_request,
            pres=pres,
        )
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with (
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
            ),
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
        ):
            px_rec_out = await self.manager.verify_pres(px_rec_in)
            save_ex.assert_called_once()

            assert px_rec_out.state == (V20PresExRecord.STATE_DONE)

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_verify_pres_indy_and_dif(self):
        pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                ),
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                ),
            ],
            will_confirm=True,
            request_presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF_REQ_NAME, ident="anoncreds"),
                AttachDecorator.data_json(DIF_PRES_REQ, ident="dif"),
            ],
        )
        pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="anoncreds",
                    format_=ATTACHMENT_FORMAT[PRES_20][
                        V20PresFormat.Format.ANONCREDS.api
                    ],
                ),
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                ),
            ],
            presentations_attach=[
                AttachDecorator.data_base64(ANONCREDS_PROOF, ident="anoncreds"),
                AttachDecorator.data_json(DIF_PRES, ident="dif"),
            ],
        )
        px_rec_in = V20PresExRecord(
            pres_request=pres_request,
            pres=pres,
        )

        self.profile.context.injector.bind_instance(
            DocumentLoader, custom_document_loader
        )
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with (
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
            ),
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
        ):
            px_rec_out = await self.manager.verify_pres(px_rec_in)
            save_ex.assert_called_once()

            assert px_rec_out.state == (V20PresExRecord.STATE_DONE)

        with (
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
            ),
            mock.patch(
                "acapy_agent.vc.vc_ld.verify.verify_presentation",
                mock.CoroutineMock(
                    return_value=PresentationVerificationResult(verified=False)
                ),
            ),
            mock.patch.object(
                AnonCredsVerifier,
                "verify_presentation",
                mock.CoroutineMock(
                    return_value=PresentationVerificationResult(verified=True)
                ),
            ),
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
        ):
            px_rec_out = await self.manager.verify_pres(px_rec_in)
            save_ex.assert_called_once()
            assert px_rec_out.state == (V20PresExRecord.STATE_DONE)
            assert px_rec_out.verified == "false"

    async def test_send_pres_ack(self):
        px_rec = V20PresExRecord()

        responder = MockResponder()
        self.profile.context.injector.bind_instance(BaseResponder, responder)

        await self.manager.send_pres_ack(px_rec)
        messages = responder.messages
        assert len(messages) == 1

        px_rec = V20PresExRecord(verified="true")

        responder = MockResponder()
        self.profile.context.injector.bind_instance(BaseResponder, responder)

        await self.manager.send_pres_ack(px_rec)
        messages = responder.messages
        assert len(messages) == 1

        px_rec = V20PresExRecord(verified="false")

        responder = MockResponder()
        self.profile.context.injector.bind_instance(BaseResponder, responder)

        await self.manager.send_pres_ack(px_rec)
        messages = responder.messages
        assert len(messages) == 1

    async def test_send_pres_ack_no_responder(self):
        px_rec = V20PresExRecord()

        self.profile.context.injector.clear_binding(BaseResponder)
        await self.manager.send_pres_ack(px_rec)

    async def test_receive_pres_ack_a(self):
        conn_record = mock.MagicMock(connection_id=CONN_ID)

        px_rec_dummy = V20PresExRecord()
        message = mock.MagicMock()

        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = px_rec_dummy
            px_rec_out = await self.manager.receive_pres_ack(message, conn_record)
            save_ex.assert_called_once()

            assert px_rec_out.state == V20PresExRecord.STATE_DONE

    async def test_receive_pres_ack_b(self):
        conn_record = mock.MagicMock(connection_id=CONN_ID)

        px_rec_dummy = V20PresExRecord()
        message = mock.MagicMock(_verification_result="true")

        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                V20PresExRecord, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = px_rec_dummy
            px_rec_out = await self.manager.receive_pres_ack(message, conn_record)
            save_ex.assert_called_once()

            assert px_rec_out.state == V20PresExRecord.STATE_DONE
            assert px_rec_out.verified == "true"

    async def test_receive_problem_report(self):
        connection_id = "connection-id"
        stored_exchange = V20PresExRecord(
            pres_ex_id="dummy-cxid",
            connection_id=connection_id,
            initiator=V20PresExRecord.INITIATOR_SELF,
            role=V20PresExRecord.ROLE_VERIFIER,
            state=V20PresExRecord.STATE_PROPOSAL_RECEIVED,
            thread_id="dummy-thid",
        )
        problem = V20PresProblemReport(
            description={
                "en": "Change of plans",
                "code": test_module.ProblemReportReason.ABANDONED.value,
            }
        )

        with (
            mock.patch.object(V20PresExRecord, "save", autospec=True) as save_ex,
            mock.patch.object(
                V20PresExRecord,
                "retrieve_by_tag_filter",
                mock.CoroutineMock(),
            ) as retrieve_ex,
            mock.patch.object(
                self.profile,
                "session",
                mock.MagicMock(return_value=self.profile.session()),
            ) as session,
        ):
            retrieve_ex.return_value = stored_exchange

            await self.manager.receive_problem_report(problem, connection_id)
            retrieve_ex.assert_called_once_with(
                session.return_value,
                {"thread_id": problem._thread_id},
                {"connection_id": connection_id},
            )
            save_ex.assert_called_once()

        assert stored_exchange.state == V20PresExRecord.STATE_ABANDONED

    async def test_receive_problem_report_x(self):
        connection_id = "connection-id"
        problem = V20PresProblemReport(
            description={
                "en": "Change of plans",
                "code": test_module.ProblemReportReason.ABANDONED.value,
            }
        )

        with mock.patch.object(
            V20PresExRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(),
        ) as retrieve_ex:
            retrieve_ex.side_effect = StorageNotFoundError("No such record")

            with self.assertRaises(StorageNotFoundError):
                await self.manager.receive_problem_report(problem, connection_id)



"""Base class for OpenAPI artifact schema."""

from marshmallow import EXCLUDE, Schema


class OpenAPISchema(Schema):
    """Schema for OpenAPI artifacts: excluding unknown fields, not raising exception."""

    class Meta:
        """OpenAPISchema metadata."""

        model_class = None
        unknown = EXCLUDE



"""Credential Issuance proof purpose class."""

from typing import TYPE_CHECKING, List

from pyld import jsonld
from pyld.jsonld import JsonLdProcessor

from ..constants import CREDENTIALS_ISSUER_URL
from ..document_loader import DocumentLoaderMethod
from ..error import LinkedDataProofException
from ..validation_result import PurposeResult
from .assertion_proof_purpose import AssertionProofPurpose

# Avoid circular dependency
if TYPE_CHECKING:
    from ..suites import _LinkedDataProof as LinkedDataProof


class CredentialIssuancePurpose(AssertionProofPurpose):
    """Credential Issuance proof purpose."""

    def validate(
        self,
        *,
        proof: dict,
        document: dict,
        suite: "LinkedDataProof",
        verification_method: dict,
        document_loader: DocumentLoaderMethod,
    ) -> PurposeResult:
        """Validate if the issuer matches the controller of the verification method."""
        try:
            result = super().validate(
                proof=proof,
                document=document,
                suite=suite,
                verification_method=verification_method,
                document_loader=document_loader,
            )

            # Return early if super check was invalid
            if not result.valid:
                return result

            # FIXME: Other implementations don't expand, but
            # if we don't expand we can't get the property using
            # the full CREDENTIALS_ISSUER_URL.
            [expanded] = jsonld.expand(
                document,
                {
                    "documentLoader": document_loader,
                },
            )

            issuer: List[dict] = JsonLdProcessor.get_values(
                expanded, CREDENTIALS_ISSUER_URL
            )

            if len(issuer) == 0:
                raise LinkedDataProofException("Credential issuer is required.")

            controller_id = result.controller.get("id")
            issuer_id = issuer[0].get("@id")

            if controller_id != issuer_id:
                raise LinkedDataProofException(
                    "Credential issuer must match the verification method controller."
                )

            return result
        except Exception as e:
            return PurposeResult(valid=False, error=str(e))



"""JSON-LD messaging Exceptions."""

from ...core.error import BaseError


class BaseJSONLDMessagingError(BaseError):
    """Base exception class for JSON-LD messaging."""


class BadJWSHeaderError(BaseJSONLDMessagingError):
    """Exception indicating invalid JWS header."""


class DroppedAttributeError(BaseJSONLDMessagingError):
    """Exception used to track that an attribute was removed."""


class MissingVerificationMethodError(BaseJSONLDMessagingError):
    """Exception indicating missing verification method from signature options."""


class SignatureTypeError(BaseJSONLDMessagingError):
    """Exception indicating Signature type error."""


class InvalidVerificationMethod(BaseJSONLDMessagingError):
    """Exception indicating an invalid verification method in doc to verify."""



"""Cache for multitenancy profiles."""

import logging
from collections import OrderedDict
from typing import Optional
from weakref import WeakValueDictionary

from ..core.profile import Profile

LOGGER = logging.getLogger(__name__)


class ProfileCache:
    """Profile cache that caches based on LRU strategy."""

    def __init__(self, capacity: int):
        """Initialize ProfileCache.

        Args:
            capacity: The capacity of the cache. If capacity is exceeded
                      profiles are closed.
        """

        LOGGER.debug(f"Profile cache initialized with capacity {capacity}")

        self._cache: OrderedDict[str, Profile] = OrderedDict()
        self.profiles: WeakValueDictionary[str, Profile] = WeakValueDictionary()
        self.capacity = capacity

    def _cleanup(self):
        """Prune cache until size matches defined capacity."""
        if len(self._cache) > self.capacity:
            LOGGER.debug(
                f"Profile limit of {self.capacity} reached."
                " Evicting least recently used profiles..."
            )
            while len(self._cache) > self.capacity:
                key, _ = self._cache.popitem(last=False)
                LOGGER.debug(f"Evicted profile with key {key}")

    def get(self, key: str) -> Optional[Profile]:
        """Get profile with associated key from cache.

        If a profile is open but has been evicted from the cache, this will
        reinsert the profile back into the cache. This prevents attempting to
        open a profile that is already open. Triggers clean up.

        Args:
            key (str): the key to get the profile for.

        Returns:
            Optional[Profile]: Profile if found in cache.

        """
        value = self.profiles.get(key)
        if value:
            if key not in self._cache:
                LOGGER.debug(
                    f"Rescuing profile {key} from eviction from cache; profile "
                    "will be reinserted into cache"
                )
                self._cache[key] = value
            self._cache.move_to_end(key)
            self._cleanup()

        return value

    def has(self, key: str) -> bool:
        """Check whether there is a profile with associated key in the cache.

        Args:
            key (str): the key to check for a profile

        Returns:
            bool: Whether the key exists in the cache

        """
        return key in self.profiles

    def put(self, key: str, value: Profile) -> None:
        """Add profile with associated key to the cache.

        If new profile exceeds the cache capacity least recently used profiles
        that are not used will be removed from the cache.

        Args:
            key (str): the key to set
            value (Profile): the profile to set
        """

        # Profiles are responsible for cleaning up after themselves when they
        # fall out of scope. Previously the cache needed to create a finalizer.
        # value.finalzer()

        # Keep track of currently opened profiles using weak references
        self.profiles[key] = value

        # Strong reference to profile to hold open until evicted
        LOGGER.debug(f"Setting profile with id {key} in profile cache")
        self._cache[key] = value

        # Refresh profile liveliness
        self._cache.move_to_end(key)
        self._cleanup()

    def remove(self, key: str):
        """Remove profile with associated key from the cache.

        Args:
            key (str): The key to remove from the cache.
        """
        del self.profiles[key]
        del self._cache[key]



"""AnonCreds Credential Offer format for v2.0 of the issue-credential protocol."""

from typing import Optional, Sequence

from marshmallow import EXCLUDE, fields

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.valid import (
    ANONCREDS_CRED_DEF_ID_EXAMPLE,
    ANONCREDS_CRED_DEF_ID_VALIDATE,
    ANONCREDS_SCHEMA_ID_EXAMPLE,
    ANONCREDS_SCHEMA_ID_VALIDATE,
    NUM_STR_WHOLE_EXAMPLE,
    NUM_STR_WHOLE_VALIDATE,
)


class AnonCredsKeyCorrectnessProof(BaseModel):
    """AnonCreds key correctness proof."""

    class Meta:
        """AnonCredsKeyCorrectnessProof metadata."""

        schema_class = "AnonCredsKeyCorrectnessProofSchema"

    def __init__(
        self,
        c: Optional[str] = None,
        xz_cap: Optional[str] = None,
        xr_cap: Sequence[Sequence[str]] = None,
        **kwargs,
    ):
        """Initialize XR cap for anoncreds key correctness proof."""
        super().__init__(**kwargs)

        self.c = c
        self.xz_cap = xz_cap
        self.xr_cap = xr_cap


class AnonCredsCorrectnessProofSchema(BaseModelSchema):
    """AnonCreds key correctness proof schema."""

    class Meta:
        """AnonCreds key correctness proof schema metadata."""

        model_class = AnonCredsKeyCorrectnessProof
        unknown = EXCLUDE

    c = fields.Str(
        required=True,
        validate=NUM_STR_WHOLE_VALIDATE,
        metadata={
            "description": "c in key correctness proof",
            "example": NUM_STR_WHOLE_EXAMPLE,
        },
    )
    xz_cap = fields.Str(
        required=True,
        validate=NUM_STR_WHOLE_VALIDATE,
        metadata={
            "description": "xz_cap in key correctness proof",
            "example": NUM_STR_WHOLE_EXAMPLE,
        },
    )
    xr_cap = fields.List(
        fields.List(
            fields.Str(
                required=True,
                metadata={
                    "description": "xr_cap component values in key correctness proof"
                },
            ),
            required=True,
            metadata={
                "description": "xr_cap components in key correctness proof",
                "many": True,
            },
        ),
        required=True,
        metadata={"description": "xr_cap in key correctness proof", "many": True},
    )


class AnonCredsCredentialOffer(BaseModel):
    """AnonCreds Credential Offer."""

    class Meta:
        """AnonCredsCredentialOffer metadata."""

        schema_class = "AnonCredsCredentialOfferSchema"

    def __init__(
        self,
        schema_id: Optional[str] = None,
        cred_def_id: Optional[str] = None,
        nonce: Optional[str] = None,
        key_correctness_proof: Optional[str] = None,
        **kwargs,
    ):
        """Initialize values ."""
        super().__init__(**kwargs)
        self.schema_id = schema_id
        self.cred_def_id = cred_def_id
        self.nonce = nonce
        self.key_correctness_proof = key_correctness_proof


class AnonCredsCredentialOfferSchema(BaseModelSchema):
    """AnonCreds Credential Offer Schema."""

    class Meta:
        """AnonCredsCredentialOffer schema metadata."""

        model_class = AnonCredsCredentialOffer
        unknown = EXCLUDE

    schema_id = fields.Str(
        required=True,
        validate=ANONCREDS_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
        },
    )

    cred_def_id = fields.Str(
        required=True,
        validate=ANONCREDS_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": ANONCREDS_CRED_DEF_ID_EXAMPLE,
        },
    )

    nonce = fields.Str(
        required=True,
        validate=NUM_STR_WHOLE_VALIDATE,
        metadata={
            "description": "Nonce in credential abstract",
            "example": NUM_STR_WHOLE_EXAMPLE,
        },
    )

    key_correctness_proof = fields.Nested(
        AnonCredsCorrectnessProofSchema(),
        required=True,
        metadata={"description": "Key correctness proof"},
    )



"""Operations supporting JWT creation and verification."""

import json
import logging
from typing import Any, Mapping, Optional, Tuple

from marshmallow import fields
from pydid import DIDUrl, Resource, VerificationMethod
from pydid.verification_method import Ed25519VerificationKey2018, Multikey

from acapy_agent.wallet.keys.manager import key_type_from_multikey, multikey_to_verkey

from ..core.profile import Profile
from ..messaging.jsonld.error import BadJWSHeaderError, InvalidVerificationMethod
from ..messaging.models.base import BaseModel, BaseModelSchema
from ..resolver.did_resolver import DIDResolver
from .base import BaseWallet
from .default_verification_key_strategy import BaseVerificationKeyStrategy
from .key_type import ED25519, KeyType
from .util import b64_to_bytes, bytes_to_b64

LOGGER = logging.getLogger(__name__)
SUPPORTED_JWT_ALGS = ("EdDSA", "ES256")


def dict_to_b64(value: Mapping[str, Any]) -> str:
    """Encode a dictionary as a b64 string."""
    return bytes_to_b64(json.dumps(value).encode(), urlsafe=True, pad=False)


def b64_to_dict(value: str) -> Mapping[str, Any]:
    """Decode a dictionary from a b64 encoded value."""
    return json.loads(b64_to_bytes(value, urlsafe=True))


def nym_to_did(value: str) -> str:
    """Return a did from nym if passed value is nym, else return value."""
    return value if value.startswith("did:") else f"did:sov:{value}"


def did_lookup_name(value: str) -> str:
    """Return the value used to lookup a DID in the wallet.

    If value is did:sov, return the unqualified value. Else, return value.
    """
    return value.split(":", 3)[2] if value.startswith("did:sov:") else value


async def jwt_sign(
    profile: Profile,
    headers: Mapping[str, Any],
    payload: Mapping[str, Any],
    did: Optional[str] = None,
    verification_method: Optional[str] = None,
) -> str:
    """Create a signed JWT given headers, payload, and signing DID or DID URL."""
    if verification_method is None:
        if did is None:
            raise ValueError("did or verificationMethod required.")

        did = nym_to_did(did)

        verkey_strat = profile.inject(BaseVerificationKeyStrategy)
        verification_method = await verkey_strat.get_verification_method_id_for_did(
            did, profile
        )
    else:
        # We look up keys by did for now
        did = DIDUrl.parse(verification_method).did
        if not did:
            raise ValueError("DID URL must be absolute")

    async with profile.session() as session:
        wallet = session.inject(BaseWallet)
        did_info = await wallet.get_local_did(did_lookup_name(did))

        header_alg = did_info.key_type.jws_algorithm
        if not header_alg:
            raise ValueError(f"DID key type '{did_info.key_type}' cannot be used for JWS")

        if not headers.get("typ", None):
            headers["typ"] = "JWT"
        headers = {
            **headers,
            "alg": header_alg,
            "kid": verification_method,
        }
        encoded_headers = dict_to_b64(headers)
        encoded_payload = dict_to_b64(payload)

        LOGGER.info(f"jwt sign: {did}")
        sig_bytes = await wallet.sign_message(
            f"{encoded_headers}.{encoded_payload}".encode(), did_info.verkey
        )

    sig = bytes_to_b64(sig_bytes, urlsafe=True, pad=False)
    return f"{encoded_headers}.{encoded_payload}.{sig}"


class JWTVerifyResult(BaseModel):
    """Result from verify."""

    class Meta:
        """JWTVerifyResult metadata."""

        schema_class = "JWTVerifyResultSchema"

    def __init__(
        self,
        headers: Mapping[str, Any],
        payload: Mapping[str, Any],
        valid: bool,
        kid: str,
    ):
        """Initialize a JWTVerifyResult instance."""
        self.headers = headers
        self.payload = payload
        self.valid = valid
        self.kid = kid


class JWTVerifyResultSchema(BaseModelSchema):
    """JWTVerifyResult schema."""

    class Meta:
        """JWTVerifyResultSchema metadata."""

        model_class = JWTVerifyResult

    headers = fields.Dict(
        required=True, metadata={"description": "Headers from verified JWT."}
    )
    payload = fields.Dict(
        required=True, metadata={"description": "Payload from verified JWT"}
    )
    valid = fields.Bool(required=True)
    kid = fields.Str(required=True, metadata={"description": "kid of signer"})
    error = fields.Str(required=False, metadata={"description": "Error text"})


async def resolve_public_key_by_kid_for_verify(
    profile: Profile, kid: str
) -> Tuple[str, KeyType]:
    """Resolve public key verkey (base58 public key) and key type from a kid."""
    resolver = profile.inject(DIDResolver)
    vmethod: Resource = await resolver.dereference(
        profile,
        kid,
    )

    if not isinstance(vmethod, VerificationMethod):
        raise InvalidVerificationMethod(
            "Dereferenced resource is not a verification method"
        )

    if isinstance(vmethod, Ed25519VerificationKey2018):
        verkey = vmethod.public_key_base58
        ktyp = ED25519
        return (verkey, ktyp)

    if isinstance(vmethod, Multikey):
        multikey = vmethod.public_key_multibase
        verkey = multikey_to_verkey(multikey)
        ktyp = key_type_from_multikey(multikey=multikey)
        return (verkey, ktyp)

    # unsupported
    raise InvalidVerificationMethod(
        f"Dereferenced method {type(vmethod).__name__} is not supported"
    )


async def jwt_verify(profile: Profile, jwt: str) -> JWTVerifyResult:
    """Verify a JWT and return the headers and payload."""
    encoded_headers, encoded_payload, encoded_signature = jwt.split(".", 3)
    headers = b64_to_dict(encoded_headers)
    if "alg" not in headers:
        raise BadJWSHeaderError("Missing 'alg' parameter in JWS header")
    if "kid" not in headers:
        raise BadJWSHeaderError("Missing 'kid' parameter in JWS header")
    if headers["alg"] not in SUPPORTED_JWT_ALGS:
        raise BadJWSHeaderError(
            f"Unsupported 'alg' value in JWS header: '{headers['alg']}'. "
            f"Supported algorithms: {', '.join(SUPPORTED_JWT_ALGS)}"
        )

    payload = b64_to_dict(encoded_payload)
    verification_method = headers["kid"]
    decoded_signature = b64_to_bytes(encoded_signature, urlsafe=True)

    async with profile.session() as session:
        (verkey, ktyp) = await resolve_public_key_by_kid_for_verify(
            profile, verification_method
        )
        wallet = session.inject(BaseWallet)
        valid = await wallet.verify_message(
            f"{encoded_headers}.{encoded_payload}".encode(),
            decoded_signature,
            from_verkey=verkey,
            key_type=ktyp,
        )

    return JWTVerifyResult(headers, payload, valid, verification_method)



from typing import Callable

import pytest

from ....messaging.models.base import BaseModelError
from ..vc_record import VCRecord

sample_json_cred_1 = """
    {
        "@context": ["https://www.w3.org/2018/credentials/v1"],
        "id": "https://eu.com/claims/DriversLicense",
        "type": ["EUDriversLicense"],
        "issuer": "did:example:123",
        "issuanceDate": "2010-01-01T19:73:24Z",
        "credentialSchema": {
            "id": "https://eu.com/claims/DriversLicense.json",
            "type": "JsonSchemaValidator2018"
        },
        "credentialSubject": {
            "id": "did:example:ebfeb1f712ebc6f1c276e12ec21",
            "accounts": [
            {
                "id": "1234567890",
                "route": "DE-9876543210"
            },
            {
                "id": "2457913570",
                "route": "DE-0753197542"
            }
            ]
        },
        "proof": {
            "type": "RsaSignature2018",
            "created": "2017-06-18T21:19:10Z",
            "proofPurpose": "assertionMethod",
            "verificationMethod": "https://example.edu/issuers/keys/1",
            "jws": "eyJhbGciOiJSUzI1NiIsImI2NCI6ZmFsc2UsImNyaXQiOlsiYjY0Il19..TCYt5XsITJX1CxPCT8yAV-TVkIEq_PbChOMqsLfRoPsnsgw5WEuts01mq-pQy7UJiN5mgRxD-WUcX16dUEMGlv50aqzpqh4Qktb3rk-BuQy72IFLOqV0G_zS245-kronKb78cPN25DGlcTwLtjPAYuNzVBAh4vGHSrQyHUdBBPM"
        }
    }
"""
sample_json_cred_2 = """
    {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1"
        ],
        "id": "http://example.edu/credentials/1872",
        "type": ["VerifiableCredential", "AlumniCredential"],
        "issuer": {
            "id": "https://example.edu/issuers/565049"
        },
        "issuanceDate": "2011-01-01T19:73:24Z",
        "credentialSchema": [
            {
                "id": "https://example.org/examples/degree.json",
                "type": "JsonSchemaValidator2018"
            }
        ],
        "credentialSubject": [
            {
                "id": "did:example:ebfeb1f712ebc6f1c276e12ec21",
                "alumniOf": {
                "id": "did:example:c276e12ec21ebfeb1f712ebc6f1",
                "name": [{
                    "value": "Example University",
                    "lang": "en"
                }, {
                    "value": "Exemple d'Université",
                    "lang": "fr"
                }]
                }
            }
        ],
        "proof": {
            "type": "RsaSignature2018",
            "created": "2017-06-18T21:19:10Z",
            "proofPurpose": "assertionMethod",
            "verificationMethod": "https://example.edu/issuers/keys/1",
            "jws": "eyJhbGciOiJSUzI1NiIsImI2NCI6ZmFsc2UsImNyaXQiOlsiYjY0Il19..TCYt5XsITJX1CxPCT8yAV-TVkIEq_PbChOMqsLfRoPsnsgw5WEuts01mq-pQy7UJiN5mgRxD-WUcX16dUEMGlv50aqzpqh4Qktb3rk-BuQy72IFLOqV0G_zS245-kronKb78cPN25DGlcTwLtjPAYuNzVBAh4vGHSrQyHUdBBPM"
        }
    }
"""
sample_json_cred_3 = """
    {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1"
        ],
        "id": "http://example.edu/credentials/1872",
        "type": ["VerifiableCredential", "AlumniCredential"],
        "issuer": {
            "id": "https://example.edu/issuers/565049"
        },
        "issuanceDate": "2010-01-01T19:73:24Z",
        "credentialSchema": "https://example.org/examples/degree.json",
        "credentialSubject": { "id": "did:example:ebfeb1f712ebc6f1c276e12ec21" },
        "proof": {
            "type": "RsaSignature2018",
            "created": "2017-06-18T21:19:10Z",
            "proofPurpose": "assertionMethod",
            "verificationMethod": "https://example.edu/issuers/keys/1",
            "jws": "eyJhbGciOiJSUzI1NiIsImI2NCI6ZmFsc2UsImNyaXQiOlsiYjY0Il19..TCYt5XsITJX1CxPCT8yAV-TVkIEq_PbChOMqsLfRoPsnsgw5WEuts01mq-pQy7UJiN5mgRxD-WUcX16dUEMGlv50aqzpqh4Qktb3rk-BuQy72IFLOqV0G_zS245-kronKb78cPN25DGlcTwLtjPAYuNzVBAh4vGHSrQyHUdBBPM"
        }
    }
"""
CONTEXTS = [
    "https://www.w3.org/2018/credentials/v1",
    "https://www.w3.org/2018/credentials/examples/v1",
]
TYPES = [
    "https://www.w3.org/2018/credentials#VerifiableCredential",
    "https://example.org/examples#UniversityDegreeCredential",
]
ISSUER_ID = "https://example.edu/issuers/14"
SUBJECT_IDS = ["did:example:ebfeb1f712ebc6f1c276e12ec21"]
SCHEMA_IDS = ["https://example.org/examples/degree.json"]
PROOF_TYPES = ["RsaSignature2018"]
GIVEN_ID = "http://example.edu/credentials/3732"
CRED_TAGS = {"tag": "value"}
CRED_VALUE = {"...": "..."}


@pytest.fixture
def record():
    def _record():
        return VCRecord(
            contexts=CONTEXTS,
            expanded_types=TYPES,
            schema_ids=SCHEMA_IDS,
            issuer_id=ISSUER_ID,
            subject_ids=SUBJECT_IDS,
            proof_types=PROOF_TYPES,
            cred_value=CRED_VALUE,
            given_id=GIVEN_ID,
            cred_tags=CRED_TAGS,
        )

    yield _record


class TestVCRecord:
    def test_create(self, record: Callable[[], VCRecord]):
        record_a = record()
        assert record_a.contexts == set(CONTEXTS)
        assert record_a.expanded_types == set(TYPES)
        assert record_a.schema_ids == set(SCHEMA_IDS)
        assert record_a.subject_ids == set(SUBJECT_IDS)
        assert record_a.proof_types == set(PROOF_TYPES)
        assert record_a.issuer_id == ISSUER_ID
        assert record_a.given_id == GIVEN_ID
        assert record_a.record_id and isinstance(record_a.record_id, str)
        assert record_a.cred_tags == CRED_TAGS
        assert record_a.cred_value == CRED_VALUE

    def test_eq(self, record: Callable[[], VCRecord]):
        record_a = record()
        record_b = record()

        assert record_a != record_b
        record_b.record_id = record_a.record_id
        assert record_a == record_b
        assert record_a != object()
        record_b.contexts.clear()
        assert record_a != record_b

    def test_serde(self, record: Callable[[], VCRecord]):
        obj = record().serialize()
        rec = VCRecord.deserialize(obj)
        assert isinstance(rec, VCRecord)

        obj_x = record()
        obj_x.cred_tags = -1  # not a dict
        with pytest.raises(BaseModelError):
            obj_x.serialize()



DID_z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL = {
    "@context": "https://www.w3.org/ns/did/v1",
    "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
    "verificationMethod": [
        {
            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
            "type": "Ed25519VerificationKey2018",
            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
            "publicKeyBase58": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
        }
    ],
    "authentication": [
        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
    ],
    "assertionMethod": [
        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
    ],
    "capabilityDelegation": [
        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
    ],
    "capabilityInvocation": [
        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
    ],
    "keyAgreement": [
        {
            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6LSbkodSr6SU2trs8VUgnrnWtSm7BAPG245ggrBmSrxbv1R",
            "type": "X25519KeyAgreementKey2019",
            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
            "publicKeyBase58": "5dTvYHaNaB7mk7iA9LqCJEHG2dGZQsvoi8WGzDRtYEf",
        }
    ],
}



"""Introduction service base classes."""

from abc import ABC, abstractmethod
from typing import Optional

from ....core.error import BaseError
from ....core.profile import ProfileSession
from .messages.invitation import Invitation


class IntroductionError(BaseError):
    """Generic introduction service error."""


class BaseIntroductionService(ABC):
    """Service handler for allowing connections to exchange invitations."""

    def __init__(self):
        """Init admin service."""

    @classmethod
    def service_handler(cls):
        """Quick accessor for conductor to use."""

        async def get_instance():
            """Return registered server."""
            return cls()

        return get_instance

    @abstractmethod
    async def start_introduction(
        self,
        init_connection_id: str,
        target_connection_id: str,
        outbound_handler,
        session: ProfileSession,
        message: Optional[str] = None,
    ):
        """Start the introduction process between two connections.

        Args:
            init_connection_id: The connection initiating the request
            target_connection_id: The connection which is asked for an invitation
            outbound_handler: The outbound handler coroutine for sending a message
            session: Profile session to use for connection, introduction records
            message: The message to use when requesting the invitation
        """

    @abstractmethod
    async def return_invitation(
        self,
        target_connection_id: str,
        invitation: Invitation,
        session: ProfileSession,
        outbound_handler,
    ):
        """Handle the forwarding of an invitation to the responder.

        Args:
            target_connection_id: The ID of the connection sending the Invitation
            invitation: The received Invitation message
            session: Profile session to use for introduction records
            outbound_handler: The outbound handler coroutine for sending a message
        """



"""Operations supporting SD-JWT creation and verification."""

import re
from typing import Any, List, Mapping, Optional, Union

from jsonpath_ng.ext import parse as jsonpath_parse
from marshmallow import fields
from sd_jwt.common import SDObj
from sd_jwt.issuer import SDJWTIssuer
from sd_jwt.verifier import SDJWTVerifier

from ..core.error import BaseError
from ..core.profile import Profile
from ..messaging.valid import StrOrDictField
from ..wallet.jwt import JWTVerifyResult, JWTVerifyResultSchema, jwt_sign, jwt_verify

CLAIMS_NEVER_SD = ["iss", "iat", "exp", "cnf"]


class SDJWTError(BaseError):
    """SD-JWT Error."""


class SDJWTIssuerACAPy(SDJWTIssuer):
    """SDJWTIssuer class for ACA-Py implementation."""

    def __init__(
        self,
        user_claims: dict,
        issuer_key,
        holder_key,
        profile: Profile,
        headers: dict,
        did: Optional[str] = None,
        verification_method: Optional[str] = None,
        add_decoy_claims: bool = False,
        serialization_format: str = "compact",
    ):
        """Initialize an SDJWTIssuerACAPy instance."""
        self._user_claims = user_claims
        self._issuer_key = issuer_key
        self._holder_key = holder_key

        self.profile = profile
        self.headers = headers
        self.did = did
        self.verification_method = verification_method

        self._add_decoy_claims = add_decoy_claims
        self._serialization_format = serialization_format
        self.ii_disclosures = []

    async def _create_signed_jws(self) -> str:
        self.serialized_sd_jwt = await jwt_sign(
            self.profile,
            self.headers,
            self.sd_jwt_payload,
            self.did,
            self.verification_method,
        )

    async def issue(self) -> str:
        """Issue an sd-jwt."""
        self._check_for_sd_claim(self._user_claims)
        self._assemble_sd_jwt_payload()
        await self._create_signed_jws()
        self._create_combined()
        return self.sd_jwt_issuance


def create_json_paths(it, current_path="", path_list=None) -> List:
    """Create a json path for each element of the payload."""
    if path_list is None:
        path_list = []

    if isinstance(it, dict):
        for k, v in it.items():
            if not k.startswith(tuple(CLAIMS_NEVER_SD)):
                new_key = f"{current_path}.{k}" if current_path else k
                path_list.append(new_key)

                if isinstance(v, dict):
                    create_json_paths(v, new_key, path_list)
                elif isinstance(v, list):
                    for i, e in enumerate(v):
                        if isinstance(e, (dict, list)):
                            create_json_paths(e, f"{new_key}[{i}]", path_list)
                        else:
                            path_list.append(f"{new_key}[{i}]")
    elif isinstance(it, list):
        for i, e in enumerate(it):
            if isinstance(e, (dict, list)):
                create_json_paths(e, f"{current_path}[{i}]", path_list)
            else:
                path_list.append(f"{current_path}[{i}]")

    return path_list


def sort_sd_list(sd_list) -> List:
    """Sorts sd_list.

    Ensures that selectively disclosable claims deepest
    in the structure are handled first.
    """
    nested_claim_sort = [(len(sd.split(".")), sd) for sd in sd_list]
    nested_claim_sort.sort(reverse=True)
    return [sd[1] for sd in nested_claim_sort]


def separate_list_splices(non_sd_list) -> List:
    """Separate list splices in the non_sd_list into individual indices.

    This is necessary in order to properly construct the inverse of
    the claims which should not be selectively disclosable in the case
    of list splices.
    """
    for item in non_sd_list:
        if ":" in item:
            split = re.split(r"\[|\]|:", item)
            for i in range(int(split[1]), int(split[2])):
                non_sd_list.append(f"{split[0]}[{i}]")
            non_sd_list.remove(item)

    return non_sd_list


def create_sd_list(payload, non_sd_list) -> List:
    """Create a list of claims which will be selectively disclosable."""
    flattened_payload = create_json_paths(payload)
    separated_non_sd_list = separate_list_splices(non_sd_list)
    sd_list = [claim for claim in flattened_payload if claim not in separated_non_sd_list]
    return sort_sd_list(sd_list)


async def sd_jwt_sign(
    profile: Profile,
    headers: Mapping[str, Any],
    payload: Mapping[str, Any],
    non_sd_list: Optional[List] = None,
    did: Optional[str] = None,
    verification_method: Optional[str] = None,
) -> str:
    """Sign sd-jwt.

    Use non_sd_list and json paths for payload elements to create a list of
    claims that can be selectively disclosable. Use this list to wrap
    selectively disclosable claims with SDObj within payload,
    create SDJWTIssuerACAPy object, and call SDJWTIssuerACAPy.issue().
    """
    non_sd_list = non_sd_list or []
    sd_list = create_sd_list(payload, non_sd_list)
    for sd in sd_list:
        jsonpath_expression = jsonpath_parse(f"$.{sd}")
        matches = jsonpath_expression.find(payload)
        if len(matches) < 1:
            raise SDJWTError(f"Claim for {sd} not found in payload.")
        else:
            for match in matches:
                if isinstance(match.context.value, list):
                    match.context.value.remove(match.value)
                    match.context.value.append(SDObj(match.value))
                else:
                    match.context.value[SDObj(str(match.path))] = match.context.value.pop(
                        str(match.path)
                    )

    return await SDJWTIssuerACAPy(
        user_claims=payload,
        issuer_key=None,
        holder_key=None,
        profile=profile,
        headers=headers,
        did=did,
        verification_method=verification_method,
    ).issue()


class SDJWTVerifyResult(JWTVerifyResult):
    """Result from verifying SD-JWT."""

    class Meta:
        """SDJWTVerifyResult metadata."""

        schema_class = "SDJWTVerifyResultSchema"

    def __init__(
        self,
        headers,
        payload,
        valid,
        kid,
        disclosures,
    ):
        """Initialize an SDJWTVerifyResult instance."""
        super().__init__(
            headers,
            payload,
            valid,
            kid,
        )
        self.disclosures = disclosures


class SDJWTVerifyResultSchema(JWTVerifyResultSchema):
    """SDJWTVerifyResult schema."""

    class Meta:
        """SDJWTVerifyResultSchema metadata."""

        model_class = SDJWTVerifyResult

    disclosures = fields.List(
        fields.List(StrOrDictField()),
        metadata={
            "description": "Disclosure arrays associated with the SD-JWT",
            "example": [
                ["fx1iT_mETjGiC-JzRARnVg", "name", "Alice"],
                [
                    "n4-t3mlh8jSS6yMIT7QHnA",
                    "street_address",
                    {"_sd": ["kLZrLK7enwfqeOzJ9-Ss88YS3mhjOAEk9lr_ix2Heng"]},
                ],
            ],
        },
    )


class SDJWTVerifierACAPy(SDJWTVerifier):
    """SDJWTVerifier class for ACA-Py implementation."""

    def __init__(
        self,
        profile: Profile,
        sd_jwt_presentation: str,
        expected_aud: Union[str, None] = None,
        expected_nonce: Union[str, None] = None,
        serialization_format: str = "compact",
    ):
        """Initialize an SDJWTVerifierACAPy instance."""
        self.profile = profile
        self.sd_jwt_presentation = sd_jwt_presentation
        self._serialization_format = serialization_format
        self.expected_aud = expected_aud
        self.expected_nonce = expected_nonce

    async def _verify_sd_jwt(self) -> SDJWTVerifyResult:
        verified = await jwt_verify(
            self.profile,
            self._unverified_input_sd_jwt,
        )
        return SDJWTVerifyResult(
            headers=verified.headers,
            payload=verified.payload,
            valid=verified.valid,
            kid=verified.kid,
            disclosures=self._disclosures_list,
        )

    async def verify(self) -> SDJWTVerifyResult:
        """Verify an sd-jwt."""
        self._parse_sd_jwt(self.sd_jwt_presentation)
        self._create_hash_mappings(self._input_disclosures)
        self._disclosures_list = list(self._hash_to_decoded_disclosure.values())

        self.verified_sd_jwt = await self._verify_sd_jwt()

        if self.expected_aud or self.expected_nonce:
            if not (self.expected_aud and self.expected_nonce):
                raise ValueError(
                    "Either both expected_aud and expected_nonce must be provided "
                    "or both must be None"
                )
            await self._verify_key_binding_jwt(
                self.expected_aud,
                self.expected_nonce,
            )
        return self.verified_sd_jwt

    async def _verify_key_binding_jwt(
        self,
        expected_aud: Union[str, None] = None,
        expected_nonce: Union[str, None] = None,
    ):
        verified_kb_jwt = await jwt_verify(
            self.profile, self._unverified_input_key_binding_jwt
        )
        self._holder_public_key_payload = self.verified_sd_jwt.payload.get("cnf", None)

        if not self._holder_public_key_payload:
            raise ValueError("No holder public key in SD-JWT")

        holder_public_key_payload_jwk = self._holder_public_key_payload.get("jwk", None)
        if not holder_public_key_payload_jwk:
            raise ValueError(
                "The holder_public_key_payload is malformed. "
                "It doesn't contain the claim jwk: "
                f"{self._holder_public_key_payload}"
            )

        if verified_kb_jwt.headers["typ"] != self.KB_JWT_TYP_HEADER:
            raise ValueError("Invalid header typ")
        if verified_kb_jwt.payload["aud"] != expected_aud:
            raise ValueError("Invalid audience")
        if verified_kb_jwt.payload["nonce"] != expected_nonce:
            raise ValueError("Invalid nonce")


async def sd_jwt_verify(
    profile: Profile,
    sd_jwt_presentation: str,
    expected_aud: Optional[str] = None,
    expected_nonce: Optional[str] = None,
) -> SDJWTVerifyResult:
    """Verify sd-jwt using SDJWTVerifierACAPy.verify()."""
    sd_jwt_verifier = SDJWTVerifierACAPy(
        profile, sd_jwt_presentation, expected_aud, expected_nonce
    )
    return await sd_jwt_verifier.verify()



"""Record instance stored and searchable by BaseStorage implementation."""

from collections import namedtuple
from typing import Optional

from uuid_utils import uuid4


class StorageRecord(namedtuple("StorageRecord", "type value tags id")):
    """Storage record class."""

    __slots__ = ()

    def __new__(cls, type, value, tags: Optional[dict] = None, id: Optional[str] = None):
        """Initialize some defaults on record."""
        if not id:
            id = uuid4().hex
        if not tags:
            tags = {}
        return super(cls, StorageRecord).__new__(cls, type, value, tags, id)



"""Core utilities and constants."""

import re

CORE_EVENT_PREFIX = "acapy::core::"
STARTUP_EVENT_TOPIC = CORE_EVENT_PREFIX + "startup"
STARTUP_EVENT_PATTERN = re.compile(f"^{STARTUP_EVENT_TOPIC}?$")
SHUTDOWN_EVENT_TOPIC = CORE_EVENT_PREFIX + "shutdown"
SHUTDOWN_EVENT_PATTERN = re.compile(f"^{SHUTDOWN_EVENT_TOPIC}?$")
WARNING_DEGRADED_FEATURES = "version-with-degraded-features"
WARNING_VERSION_MISMATCH = "fields-ignored-due-to-version-mismatch"
WARNING_VERSION_NOT_SUPPORTED = "version-not-supported"



"""did:jwk: resolver implementation."""

import re
from typing import Optional, Pattern, Sequence, Text

from acapy_agent.config.injection_context import InjectionContext
from acapy_agent.core.profile import Profile
from acapy_agent.resolver.base import BaseDIDResolver, ResolverError, ResolverType
from acapy_agent.wallet.jwt import b64_to_dict


class JwkDIDResolver(BaseDIDResolver):
    """did:jwk: resolver implementation."""

    PATTERN = re.compile(r"^did:jwk:(?P<did>.*)$")

    def __init__(self):
        """Initialize the resolver."""
        super().__init__(ResolverType.NATIVE)

    async def setup(self, context: InjectionContext):
        """Perform required setup for the resolver."""
        pass

    @property
    def supported_did_regex(self) -> Pattern:
        """Return supported DID regex."""
        return self.PATTERN

    async def _resolve(
        self,
        profile: Profile,
        did: str,
        service_accept: Optional[Sequence[Text]] = None,
    ) -> dict:
        """Resolve a DID."""
        if match := self.PATTERN.match(did):
            encoded = match.group("did")
        else:
            raise ResolverError(f"Invalid DID: {did}")

        jwk = b64_to_dict(encoded)
        doc = {
            "@context": [
                "https://www.w3.org/ns/did/v1",
                "https://w3id.org/security/suites/jws-2020/v1",
            ],
            "id": f"did:jwk:{encoded}",
            "verificationMethod": [
                {
                    "id": f"did:jwk:{encoded}#0",
                    "type": "JsonWebKey2020",
                    "controller": f"did:jwk:{encoded}",
                    "publicKeyJwk": jwk,
                }
            ],
        }

        use = jwk.get("use")
        if use == "sig":
            doc.update(
                {
                    "assertionMethod": [f"did:jwk:{encoded}#0"],
                    "authentication": [f"did:jwk:{encoded}#0"],
                    "capabilityInvocation": [f"did:jwk:{encoded}#0"],
                    "capabilityDelegation": [f"did:jwk:{encoded}#0"],
                }
            )
        elif use == "enc":
            doc.update({"keyAgreement": [f"did:jwk:{encoded}#0"]})

        return doc
